/* $Id: authdb_dir.C,v 1.8 2002/08/21 13:54:47 max Exp $ */

/*
 *
 * Copyright (C) 2001 David Mazieres (dm@uun.org)
 * Copyright (C) 2002 Kevin Fu (fubob@mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "authdb.h"
#include "qhash.h"
#include "rxx.h"
#include "parseopt.h"
#include "sfsmisc.h"
#include <grp.h>
#include "sfscrypt.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fts.h>
#include <stdio.h>

static rxx colon (":");

struct authcursor_dir : public authcursor {
  const str dirpath;
  const str cachepath;
  u_int32_t uidstart;
  u_int32_t uidend;
  u_int32_t gid;
  int depth;

  ptr<lockfile> lf;
  mode_t perm;
  FTS *f;
  FTSENT *e;
  bool error;

  authcursor_dir (str p, str c, u_int32_t uidstart,
		  u_int32_t uidend, u_int32_t gid, int depth,
		  mode_t pm, ptr<lockfile> l)
    : dirpath (p), cachepath (c), uidstart (uidstart), uidend (uidend),
      gid (gid), depth (depth), lf (l), perm (pm), f (NULL), e (NULL),
      error (false) 
  { reset (); }
  ~authcursor_dir ();
  void reset ();
  bool next ();

  bool find_user_name (str name) { return false; }
  bool find_user_uid (u_int32_t uid) { return false; }
  bool find_group_name (str name) { return false; }
  bool find_group_gid (u_int32_t gid) { return false; }
  void getgroups (vec<u_int32_t> *groups, str user)
  { groups->push_back (gid); }

  bool find_user_pubkey (const sfspub &pk);
  bool cache_read (const sfspub &pk, u_int32_t *uid);
  bool allocate_id (str filename, u_int32_t *uid);

};

authcursor_dir::~authcursor_dir ()
{
  fts_close (f);
}

void
authcursor_dir::reset ()
{
  if (f != NULL)
    fts_close (f);

  char *paths[2];
  paths[0] = const_cast<char *> (dirpath.cstr ());
  paths[1] = NULL;

  f = fts_open (paths, FTS_LOGICAL, NULL);
  if (f == NULL) {
    if (errno != ENOENT)
      warn ("%s: %m\n", dirpath.cstr ());
  }
}

/*
   The format of the cache file is:
     uid:filename
   where the file named by "filename" contains just a public key.
*/
static bool
str2idfilename (u_int32_t *id, str &filename, str s)
{
  vec<str> uv;
  if (split (&uv, colon, s, 2, true) != 2)
    return false;

  if (!convertint (uv[0], id))
    return false;
  
  filename = uv[1];
  return true;
}


/* if line contains a public key already cached, return the
   saved authentry 
 */
bool
authcursor_dir::cache_read (const sfspub &pk, u_int32_t *uid)
{
  if (pk.get_pubkey_nbits () < sfs_minpubkeysize)
    return false;

  int fd = ::open (cachepath, O_RDONLY);
  if (fd < 0) 
    return false;

  str filename, pubkeytest;
  u_int32_t id;
  suio buf;
  buf.clear ();
  int lineno = 1;
  bool flush = false;
  str line;
  for (;;) {
    while (!(line = suio_getline (&buf))) {
      if (buf.resid () > 0x20000) {
	buf.clear ();
	flush = true;
	continue;
      }
      int n = buf.input (fd);
      if (n > 0)
	continue;
      if (n < 0)
	warn << cachepath << ": " << strerror (errno) << "\n";
      else if (buf.resid ())
	warn << cachepath << ": " << lineno << ": incomplete last line\n";
      close (fd);
      return false;
    }
    lineno++;
    if (flush)
      warn << cachepath << ": " << lineno-1 << ": line too long\n";
    else if (!str2idfilename (&id, filename, line)) {
      warn << cachepath << ": " << lineno-1 << ": syntax error\n";
      continue;
    }

    pubkeytest = file2str (filename);
    ptr<sfspub> tpk = sfscrpt.alloc (pubkeytest);
    if (tpk && *tpk == pk) {
      *uid = id;
      close (fd);
      return true;
    }
  }
}

/*
  format of file is
    uid1:filenamen
    ...
    uidn:filenamen
  where it's in ascending order by uidi
*/
bool
authcursor_dir::allocate_id (str filename, u_int32_t *uid)
{
  if (!lf->ok ()) {
    warn << "authcursor_dir::allocate_id: lockfile "
	 << cachepath << " molested\n";
    return false;
  }

  FILE *F = fopen (cachepath, "r");

  if (!F) {
    if (errno != ENOENT)
      return false;
    else
      *uid = uidstart;
  }
  else {
    // find the first open uid
    *uid = uidstart;
    char linebuf[512];
    // XXX will miscount if a PK is more than 512 ascii chars... 
    while (fgets (linebuf, sizeof (linebuf), F))
      (*uid)++;

    fclose (F);
    if (*uid > uidend) {
      warn << cachepath << ": no uids left\n";
      return false;
    }
  }

  int fd = ::open (cachepath, O_WRONLY|O_CREAT|O_APPEND, perm);
  if (fd < 0) {
    warn << cachepath << ": " << strerror (errno) << "\n";
    return false;
  }

  char intbuf[sizeof(u_int32_t)];
  snprintf (intbuf, sizeof(u_int32_t)+1, "%u", *uid);
  str line = strbuf () << intbuf << ":" << filename << "\n";

  if ((write (fd, line.cstr (), line.len ())) != (int) line.len ()) {
    warn << cachepath << ": linewrite " << strerror (errno) << "\n";
    close (fd);
    return false;
  }
  close (fd);
  
  return true;
}


bool
authcursor_dir::next ()
{
  if (!f) {
    warn ("%s: %m\n", dirpath.cstr ());
    return false;
  }

  e = fts_read (f);

  while (e != NULL && e->fts_info != FTS_F) {
    // depth == -1 means not to limit the depth of the search 
    if (depth != -1 && e->fts_level >= depth) 
      fts_set (f, e, FTS_SKIP);
    e = fts_read (f);
  }

  if (e == NULL)
    return false;

  ae.set_type (SFSAUTH_USER);
  str pk = file2str (e->fts_path);
  if (pk) pubkey =  sfscrpt.alloc (pk);
  else pubkey = NULL;
  return true;
}

bool
authcursor_dir::find_user_pubkey (const sfspub &pk)
{
  reset ();
  while (next ()) {
    if (!pubkey) 
      continue;
    // XXX - was comparing key to NULL.  Meaning?
    if (pubkey->get_pubkey_nbits () > sfs_minpubkeysize && pk == *pubkey) {
      assert (e);
      
      str filename = strbuf () << str (e->fts_path, e->fts_pathlen);
      if (!cache_read (*pubkey, &ae.userinfo->id) &&
	  !allocate_id (filename, &ae.userinfo->id)) {
	warn << "authcursor_dir::find_user_pubkey: unable to add auth user\n";
	return false;
      }
      
      sfsauth_userinfo *ui = ae.userinfo;
      char intbuf[sizeof(u_int32_t)];
      snprintf (intbuf, sizeof(u_int32_t)+1, "%u", ui->id);
      ui->name = intbuf;
      ui->vers = 0;
      ui->gid = gid;
      ui->owner.clear ();
      ui->privs = "";
      ui->pwauth = "";
      ui->audit = "";
      return true;
    }
  }
  return false;
}

ptr<authcursor>
authdb_dir::open (bool writable, mode_t perm, bool wait = false)
{
  if (writable)
    return NULL;

  return New refcounted<authcursor_dir> (dirpath, cachepath, uidstart, 
					 uidend, gid, depth, perm, lf);
}


