/* $Id: sfscrypt.h,v 1.2 2002/08/21 13:54:48 max Exp $ */
/*
 *
 * Copyright (C) 2002 David Mazieres 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

// SFSCRYPT
//
// This library servers as an abstract interface between specific
// cryptographic implementations and the generic XDR implementations
// of keys, ciphertexts, and signatures.
//

// XXX - a lot of functions should be inlined.... if we can get the compiler
// to agree with that..

#ifndef _SFSMISC_SFSCRYPT_H
#define _SFSMISC_SFSCRYPT_H 1

#include "crypt.h"
#include "rabin.h"
#include "serial.h"
#include "sfs_prot.h"
#include "sfsauth_prot.h"
#include "sfsagent.h"

#define SFS_ENCRYPT   (1 << 0)
#define SFS_VERIFY    (1 << 1)
#define SFS_DECRYPT   (1 << 2)
#define SFS_SIGN      (1 << 3)

class sfspub {
public:
  sfspub (sfs_keytype kt, u_char o = 0) : ktype (kt), opts (o) {}
  virtual ~sfspub () {}

  // Should be implemented by child classes
  virtual bool export_pubkey (sfs_pubkey2 *k) const = 0;
  virtual bool export_pubkey (strbuf &sb) const = 0;
  virtual size_t get_pubkey_nbits () const = 0;
  virtual bool operator== (const sfs_pubkey2 &pk) const { return false; }
  virtual bool encrypt (sfs_ctext2 *ct, const str &msg) { return false; }
  virtual bool verify (const sfs_sig2 &sig, const str &msg) { return false; }

  // Have reasonable default behavior
  virtual bool operator== (const str &s) const;
  virtual bool operator== (const sfspub &p) const;

  // Legacy Code -- To be Phased out
  virtual bool is_v2 () { return true; }
  virtual bool export_pubkey (sfs_pubkey *k) const { return false; }
  virtual bool encrypt (sfs_ctext *ct, const str &msg) { return false; }
  virtual bool verify_r (const bigint &n, size_t len, str &msg) 
  { return false; }

  bool check_opts () { return (!get_opt (get_bad_opts ())); }
protected:
  bool verify_init (const sfs_sig2 &sig, const str &msg) const;
  virtual u_char get_bad_opts () { return (SFS_DECRYPT | SFS_SIGN); }
  bool get_opt (u_char o) { return (opts & o); }
  const sfs_keytype ktype;
  const u_char opts;
};

class sfspriv : virtual public sfspub {
public:
  sfspriv (sfs_keytype kt, u_char o = 0) : sfspub (kt, o) {}
  virtual ~sfspriv () {}

  // The Following  5 functions should be implemented by child classes
  virtual bool export_privkey (sfs_privkey2_clear *k) const = 0;
  virtual bool export_privkey (str *s, const eksblowfish *eksb) const = 0;
  virtual bool get_privkey_hash (u_int8_t *buf, const sfs_hash &hostid) 
    const = 0;
  virtual bool decrypt (const sfs_ctext2 &ct, str *msg) ;
  virtual bool sign (sfs_sig2 *sig, const str &msg) { return false; }
  virtual bool load_ctext (const sfs_ctext2 &ct) { return false; }
  virtual bool decrypt (str *msg) { return false; }

  // Legacy Code -- to be phased out
  virtual bool decrypt (const sfs_ctext &ct, str *msg) ;
  virtual bool load_ctext (const sfs_ctext &ct) { return false; }
  virtual bool sign_r (sfs_sig *sig, const str &msg) { return false; }

  // Should be implemented in terms of export_privkey (eksblowfish)
  virtual bool export_privkey (str *s, const str &kn, str pwd, u_int cost) 
    const = 0;

protected:
  u_char get_bad_modes () { return 0; }
};

class sfsca { // SFS Crypt Allocator
public:
  sfsca () : ktype (SFS_NOKEY) {}
  sfsca (sfs_keytype x, const str &s) : ktype (x), skey (s) {} 

  // These  5 functions should be implemented by child classes
  virtual ptr<sfspub>  alloc (const sfs_pubkey2 &pk, u_char o = 0) const = 0;
  virtual ptr<sfspub>  alloc (const str &s, u_char o = 0) const = 0;
  virtual ptr<sfspriv> alloc (const sfs_privkey2_clear &pk, u_char o = 0) 
    const = 0;
  virtual ptr<sfspriv> alloc (u_int nbits, u_char o = 0) const = 0;
  virtual ptr<sfspriv> alloc_priv (const str &sk, u_char o = 0) const = 0;

  // Legacy Code
  virtual ptr<sfspub>  alloc (const sfs_pubkey &pk, u_char o = 0) const 
  { return NULL; }

  // Should be implemented in terms of alloc_priv ..
  virtual ptr<sfspriv> alloc (str ske, str *kn, str *pwd, u_int *cost,
			      u_char o = 0) const = 0;
  virtual ptr<sfspriv> alloc (const str &esk, const eksblowfish *eksb, 
			      u_char o = 0) const = 0;

  virtual ~sfsca () {}
  
  ihash_entry<sfsca> hlinks;
  ihash_entry<sfsca> hlinkx;
  sfs_keytype ktype;
  str skey;
};


class sfs_rabin_pub : virtual public sfspub {
public:
  sfs_rabin_pub (ref<rabin_pub> k, u_char o = 0)
    : sfspub (SFS_RABIN, o), pubk (k) {}
  bool verify (const sfs_sig2 &sig, const str &msg);
  bool export_pubkey (sfs_pubkey2 *k) const;
  bool export_pubkey (strbuf &b) const;
  bool operator== (const sfs_pubkey2 &pk) const;
  bool operator== (const str &s) const;
  size_t get_pubkey_nbits () const;
  bool encrypt (sfs_ctext2 *ct, const str &msg);

  // Legacy Code
  bool verify_r (const bigint &n, size_t len, str &msg);
  bool export_pubkey (sfs_pubkey *k) const;
  bool encrypt (sfs_ctext *ct, const str &msg);
  bool is_v2 () { return false; }
private:
  ref<rabin_pub> pubk;
};

class sfs_rabin_priv : public sfs_rabin_pub, public sfspriv {
public:
  sfs_rabin_priv (ref<rabin_priv> k, u_char o = 0)
    : sfspub (SFS_RABIN, o), sfs_rabin_pub (k), 
      sfspriv (SFS_RABIN, o), privk (k) {}
  bool sign (sfs_sig2 *s, const str &msg);
  bool export_privkey (sfs_privkey2_clear *k) const;
  bool export_privkey (str *s, const str &kn, str pwd, u_int cost) const;
  bool export_privkey (str *s, const eksblowfish *eksb) const;
  bool load_ctext (const sfs_ctext2 &ct);
  bool decrypt (str *msg);
  bool get_privkey_hash (u_int8_t *buf, const sfs_hash &hostid) const;

  // Legacy Code
  bool load_ctext (const sfs_ctext &ct);
  bool sign_r (sfs_sig *sig, const str &msg);

private:
  ref<rabin_priv> privk;
  bigint ctext;
};

class sfs_rabin_alloc : public sfsca {
public:
  sfs_rabin_alloc () : sfsca (SFS_RABIN, "rabin") {}
  sfs_rabin_alloc (const str &s) : sfsca (SFS_RABIN, s) {}

  ptr<sfspub>  alloc (const sfs_pubkey2 &pk, u_char o = 0) const;
  ptr<sfspub>  alloc (const str &s, u_char o = 0) const;
  ptr<sfspriv> alloc (const sfs_privkey2_clear &pk, u_char o = 0) const;
  ptr<sfspriv> alloc (str ske, str *kn, str *pwd, u_int *cost, u_char o = 0)
    const;
  ptr<sfspriv> alloc (const str &esk, const eksblowfish *eksb, u_char o = 0)
    const;
  ptr<sfspriv> alloc (u_int nbits, u_char o = 0) const;
  ptr<sfspriv> alloc_priv (const str &sk, u_char o = 0) const;

  // Old protocol
  ptr<sfspub> alloc (const sfs_pubkey &pk, u_char o = 0) const;
};

class sfscrypt {
public:
  ptr<sfspub>  alloc (const sfs_pubkey2 &pk, u_char o = 0) const;
  ptr<sfspub>  alloc (const str &s, u_char o = 0) const;
  ptr<sfspriv> alloc (const sfs_privkey2_clear &pk, u_char o = 0) const;
  ptr<sfspriv> alloc (const str &ske, str *kn, str *pwd, u_int *cost, 
		      u_char o = 0) const;
  ptr<sfspriv> alloc (sfs_keytype xt, const str &esk, const eksblowfish *eksb, 
		      u_char o = 0) const;
  ptr<sfspriv> alloc (sfs_keytype xt, u_int nbits, u_char o = 0) const;
  ptr<sfspriv> alloc_priv (const str &sk, u_char o = 0) const;

  // Legacy Code
  ptr<sfspub>  alloc (const sfs_pubkey &pk, u_char o = 0) const;

  bool verify (const sfs_pubkey2 &pk, const sfs_sig2 &sig, const str &msg) 
    const;

  void add (sfsca *c)
  {
    if (c->skey.len ()) { strtab.insert (c); }
    if (c->ktype && !xttab[c->ktype]) { xttab.insert (c); }
  }

  sfscrypt::sfscrypt () 
  {
    // Add Rabin to Table
    add (New sfs_rabin_alloc ());
    add (New sfs_rabin_alloc ("SK1"));
  }

private:
  ihash<str, sfsca, &sfsca::skey, &sfsca::hlinks> strtab; 
  ihash<sfs_keytype, sfsca, &sfsca::ktype, &sfsca::hlinkx> xttab; 
};

extern sfscrypt sfscrpt;

#endif /* _SFSMISC_SFSCRYPT_H */
