/*
 *
 * Copyright (C) 2001  Frank Dabek (fdabek@lcs.mit.edu), 
 *                     Frans Kaashoek (kaashoek@lcs.mit.edu),
 *   		       Massachusetts Institute of Technology
 * 
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
#include "gcc_compat.h"
#include <arpc.h>

#include <chord.h>

#include "dhash_common.h"
#include "dhash_impl.h"
#include "dhashcli.h"
#include "verify.h"
#include "keyhash_idx.h"

#include <auditserver.h>
#include <blacklist.h>
#include <merkle.h>
#include <merkle_server.h>
#include <merkle_misc.h>
#include <dhash_prot.h>
#include <dhash_store.h>
#include <chord.h>
#include <chord_types.h>
#include <comm.h>
#include <location.h>
#include <locationtable.h>
#include <misc_utils.h>
#include <ida.h>
#include <id_utils.h>
#include <rpctypes.h>
#include <sfscrypt.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <modlogger.h>
#define warning modlogger ("dhash", modlogger::WARNING)
#define info  modlogger ("dhash", modlogger::INFO)
#define trace modlogger ("dhash", modlogger::TRACE)

#include <merkle_sync_prot.h>
#include <configurator.h>

struct dhash_config_init {
  dhash_config_init ();
} dci;

dhash_config_init::dhash_config_init ()
{
  bool ok = true;

#define set_int Configurator::only ().set_int
  /** MTU **/
  ok = ok && set_int ("dhash.mtu", 1210);
  ok = ok && set_int ("dhash.sync_timer", 30); // db sync timer
  
  ok = ok && set_int ("dhash.ch_maxreps", 5);  // total fragments
  ok = ok && set_int ("dhash.dfrags", 2);      // # of frags for construction
  ok = ok && set_int ("dhash.ch_minreps", 3);  // min of alive frags
  ok = ok && set_int ("dhash.ch_reptimeout", 1 * 60 * 60);
  ok = ok && set_int ("dhash.replicas", 5);
  ok = ok && set_int ("dhash.pk_minreps", 3);
  ok = ok && set_int ("dhash.pk_reptimeout", 1 * 60 * 60);
  
  // Josh magic....
  ok = ok && set_int ("dhash.missing_outstanding_max", 15);
  
  ok = ok && set_int ("merkle.keyhash_offer_timer", 10);
  ok = ok && set_int ("merkle.keyhash_timer", 60);
  ok = ok && set_int ("merkle.replica_timer", 5*60);
  ok = ok && set_int ("merkle.prt_timer", 5);

  assert (ok);
#undef set_int
}



// Things that read from Configurator
#define DECL_CONFIG_METHOD(name,key)			\
u_long							\
dhash::name ()						\
{							\
  static bool initialized = false;			\
  static int v = 0;					\
  if (!initialized) {					\
    initialized = Configurator::only ().get_int (key, v);	\
    assert (initialized);				\
  }							\
  return v;						\
}

DECL_CONFIG_METHOD(reptm, "merkle.replica_timer")
DECL_CONFIG_METHOD(keyhashtm, "merkle.keyhash_timer")
DECL_CONFIG_METHOD(keyhashofftm, "merkle.keyhash_offer_timer")
DECL_CONFIG_METHOD(synctm, "dhash.sync_timer")
DECL_CONFIG_METHOD(num_dfrags, "dhash.dfrags")
DECL_CONFIG_METHOD(dhash_mtu, "dhash.mtu")
DECL_CONFIG_METHOD(ch_minreps, "dhash.ch_minreps")
DECL_CONFIG_METHOD(ch_maxreps, "dhash.ch_maxreps")
DECL_CONFIG_METHOD(ch_reptimeout, "dhash.ch_reptimeout")
DECL_CONFIG_METHOD(pk_minreps, "dhash.pk_minreps")
DECL_CONFIG_METHOD(pk_maxreps, "dhash.replicas")
DECL_CONFIG_METHOD(pk_reptimeout, "dhash.pk_reptimeout")
#undef DECL_CONFIG_METHOD

// Pure virtual destructors still need definitions
dhash::~dhash () {}

ref<dhash>
dhash::produce_dhash (str dbname, str auditsock)
{
  return New refcounted<dhash_impl> (dbname, auditsock);
}

dhash_impl::~dhash_impl ()
{
  if (pmaint_obj) {
    delete pmaint_obj;
    pmaint_obj = NULL;
  }
  if (msrv) {
    delete msrv;
    msrv = NULL;
  }
  if (cli) {
    delete cli;
    cli = NULL;
  }
}

void
dhash_impl::check_avg_rtt() {
  float sum_lat = 0;
  float avg_lat = 0;
  float sum_rpcs = 0;
  float avg_rpcs = 0;
  ptr<location> loc = host_node->locations->next_loc_circular(host_node->my_ID());
  
  for(unsigned i=0; i<host_node->locations->size(); i++) {
    sum_lat += loc->clear_latency();
    sum_rpcs += loc->recent_nrpc();
    loc->clear_recent_nrpcs();
    loc = host_node->locations->next_loc_circular(loc->id());
  }

  avg_lat = sum_lat / (host_node->locations->size());
  avg_rpcs = sum_rpcs / (host_node->locations->size());

#if 1
  extern cbv register_accept;
  extern cbv unregister_accept;

  if (avg_lat > 200000 && unregister_accept) {
    (unregister_accept)();
  } else {
    (register_accept)();
  }
#endif

  delaycb(2, wrap(this, &dhash_impl::check_avg_rtt));
}

void 
dhash_impl::print_locations() {
  warn << "Location Table\n";

  ptr<location> loc = host_node->locations->next_loc_circular(host_node->my_ID());
  
  for(unsigned i=0; i<host_node->locations->size(); i++) {
    warn << "  " << loc->alive()
	 << "  " << loc->id() 
	 << " (" << armor64A(loc->id()) 
	 << ") "  
	 << (int) loc->distance()
	 << " "
	 << (int) loc->recent_nrpc()
	 << "\n";
    loc = host_node->locations->next_loc_circular(loc->id());
  }

  delaycb(60, wrap(this, &dhash_impl::print_locations));
}


dhash_impl::dhash_impl (str dbname, str auditsock) :
  missing_outstanding (0),
  pk_partial_cookie (1),
  host_node (NULL),
  cli (NULL),
  msrv (NULL),
  pmaint_obj (NULL),
  replica_syncer (NULL),
  keyhash_mgr_rpcs (0),
  keyhash_off_rpcs (0),
  keyhash_off_errs (0),
  merkle_rep_tcb (NULL),
  keyhash_mgr_tcb (NULL),
  bytes_stored (0),
  keys_stored (0),
  keys_cached (0),
  keys_others (0),
  bytes_served (0),
  keys_served (0),
  rpc_answered (0)
{
  storage = New refcounted<dhash_blockstore>();
  storage->opendb(dbname, true, synctm());

  kidx = New refcounted<keyhash_idx> (storage);
  auditserver::opensocket(auditsock, kidx);

  dhcs = strbuf () << dbname;
}

void
dhash_impl::init_after_chord (ptr<vnode> node)
{
  host_node = node;
  assert (host_node);

  // merkle state
  msrv = New merkle_server (storage, 
			    wrap (node, &vnode::addHandler),
			    wrap (this, &dhash_impl::missing),
			    host_node);

  assert(ch_minreps() >= num_dfrags() - 1);
  assert(ch_maxreps() >= ch_minreps());

  // RPC demux
  trace << host_node->my_ID () << " registered dhash_program_1\n";
  host_node->addHandler (dhash_program_1, wrap(this, &dhash_impl::dispatch));

  // the client helper class (will use for get_key etc)
  cli = New dhashcli (node, dhcs);

  // Replication agents
  merkle_rep_tcb = delaycb
    (reptm ()+(random () % reptm ()),
     wrap (this, &dhash_impl::replica_maintenance_timer, 0));
  keyhash_mgr_tcb = delaycb
    (keyhashtm ()+(random () % keyhashtm ()),
     wrap (this, &dhash_impl::keyhash_mgr_timer));

  // Migration agents
  pk_migrator = New pk_migrate (cli, host_node, storage, kidx, keyhashofftm(),
				pk_minreps(), pk_maxreps(), pk_reptimeout());
  //pmaint_obj = New pmaint (cli, host_node, storage, wrap (this, &dhash_impl::db_delete_immutable));
  if (1) 
    delaycb(60, wrap(this, &dhash_impl::print_locations));

  delaycb(2, wrap(this, &dhash_impl::check_avg_rtt));
}

void
dhash_impl::missing (ptr<location> from, bigint key)
{
  // throttle the block downloads
  if (missing_outstanding > MISSING_OUTSTANDING_MAX) {
    if (missing_q[key] == NULL) {
      missing_state *ms = New missing_state (key, from);
      missing_q.insert (ms);
    }
    return;
  }

  warn << "merkle: missing key "  << key << ", fetching\n";

  // calculate key range that we should be storing
  vec<ptr<location> > preds = host_node->preds ();
  assert (preds.size () > 0);

  missing_outstanding++;
  assert (missing_outstanding >= 0);
  cli->retrieve (blockID (key, DHASH_CONTENTHASH, DHASH_FRAG),
		 wrap (this, &dhash_impl::missing_retrieve_cb, key),
		 // XXX this is wrong, should be projkey
		 key);
}

void
dhash_impl::missing_retrieve_cb (bigint key, dhash_stat err,
				 ptr<dhash_block> b, route r)
{
  missing_outstanding--;
  assert (missing_outstanding >= 0);

  if (err)
    trace << "Could not retrieve key " << key << "\n";
  else {
    assert (b);
    // Oh, the memory copies.
    str blk (b->data, b->len);
    u_long m = Ida::optimal_dfrag (b->len, dhash::dhash_mtu ());
    if (m > num_dfrags ())
      m = num_dfrags ();
    str frag = Ida::gen_frag (m, blk);
    ref<dbrec> d = New refcounted<dbrec> (frag.cstr (), frag.len ());
    ref<dbrec> k = id2dbrec (key);

    bigint projkey = extractpid(b->data, b->len);
    if (blacklist::only().deny_project(host_node->my_location()->address(), projkey)) {
      // XXX
      warn
	<< "Missing block, but unwilling to store. (just wasted bandwidth)\n";

    } else {
      int ret = db_insert_immutable (projkey, k, d, DHASH_CONTENTHASH);
      if (ret != 0)
	warning << "merkle db_insert_immutable failure: "
		<< db_strerror (ret) << "\n";
    }
  }

  while ((missing_outstanding <= MISSING_OUTSTANDING_MAX)
	 && (missing_q.size () > 0)) {
    missing_state *ms = missing_q.first ();
    assert (ms);
    missing_q.remove (ms);
    missing (ms->from, ms->key);
    delete ms;
  }
}

#define MAX_PKS_IN_RPC 256

void
dhash_impl::keyhash_sendblk_succ_done (dhash_stat err, bool present)
{
  keyhash_mgr_rpcs --;
}

void
dhash_impl::keyhash_offerpks_succ_done
  (ptr<location> dest, ref<dhash_offer_pks_res> res, clnt_stat err)
{
  keyhash_mgr_rpcs --;
  if (!err && res && res->status == DHASH_OK) {
    for (unsigned i=0; i<res->resok->keys.size (); i++) {
      keyhash_mgr_rpcs ++;
      bigint projkey;
      ptr<dbrec> d = kidx->lookup (res->resok->keys [i], projkey);
      cli->sendblock
	(dest, blockID (res->resok->keys [i], DHASH_KEYHASH, DHASH_BLOCK),
	 projkey, d, wrap (this, &dhash_impl::keyhash_sendblk_succ_done));
    }
    if (res->resok->keys.size () > 0) {
      warn << "offer (for replicas) accepted, send "
           << res->resok->keys.size ()
           << " blocks to host " << dest->id () << "\n";
    }
  }
  else
    warn << "offer (for replicas) rejected\n";
}

void
dhash_impl::offer_keyhash_blocks (ptr<location> dest, chordID rngmin, chordID rngmax) 
{
  // this code will only offer keys if <dest> is willing to store them.
  ptr<keyhash_idx::offer_state> s = kidx->findkeys (rngmin, rngmax);

  if (!s->keys.size ())
    return;
  assert (s->vers.size () == s->keys.size ());

  unsigned seen_keys = 0;
  while(seen_keys < s->keys.size()) {

    // Find out which keys <dest> is willing to take.
    vec<int> current_rpc; // indexes into s->keys[]
    while( current_rpc.size() < MAX_PKS_IN_RPC && seen_keys < s->keys.size() ) {
      bigint projkey;
      kidx->getprojkey(s->keys[seen_keys], projkey);

      if ( !blacklist::only().deny_project(dest->address(), projkey) )
	current_rpc.push_back(seen_keys);
      seen_keys++;
    }
    
    ref<dhash_offer_pks_res> res =
      New refcounted<dhash_offer_pks_res> (DHASH_OK);
    ref<s_dhash_offer_pks_arg> arg =
      New refcounted<s_dhash_offer_pks_arg> ();

    arg->keys.setsize(current_rpc.size());
    arg->versions.setsize(current_rpc.size());

    for (unsigned i=0; i<current_rpc.size(); i++) {
      arg->keys[i] = s->keys[ current_rpc[i] ];
      arg->versions[i] = s->vers[ current_rpc[i] ];
    }

    keyhash_mgr_rpcs ++;
    host_node->doRPC
      (dest, dhash_program_1, DHASHPROC_OFFER_PKS, arg, res,
       wrap (this, &dhash_impl::keyhash_offerpks_succ_done,
	     dest, res));
  }
}

void
dhash_impl::keyhash_mgr_timer ()
{
  keyhash_mgr_tcb = NULL;

  if (keyhash_mgr_rpcs == 0) {
    vec<ptr<location> > replicas;
    vec<chordID> range_min, range_max;
    
    get_pred_replicas (host_node, pk_minreps(), pk_maxreps(), 
		       pk_reptimeout(), replicas, range_min, range_max);
    get_succ_replicas (host_node, host_node->my_ID(), 
	               pk_minreps(), pk_maxreps(), 
		       pk_reptimeout(), replicas, range_min, range_max, false);

    for(unsigned i=0; i<replicas.size(); i++) {
      if (0) warn << "  replica " << replicas[i]->id() << " from " 
		  << range_min[i] << " to " << range_max[i] << "\n";
      offer_keyhash_blocks (replicas[i], range_min[i], range_max[i]);
    }
  } 
  else
    warn << "still " << keyhash_mgr_rpcs << " outstanding mgr rpcs\n";
  
  keyhash_mgr_tcb =
    delaycb (keyhashtm (), wrap (this, &dhash_impl::keyhash_mgr_timer));
}

// using succ list
void
dhash_impl::replica_maintenance_timer (u_int i)
{
  bigint rngmin, rngmax;
  vec<chordID> range_min, range_max;
  ptr<merkle_tree> specific_tree;
  vec<bigint> exclu_proj;

  vec<ptr<location> > replicas;
  get_succ_replicas (host_node, host_node->my_ID(), 
                     ch_minreps(), ch_maxreps(), ch_reptimeout(), 
		     replicas, range_min, range_max, false);

  if (i >= replicas.size())
    i = 0;
  if (replicas.size() == 0)
    goto out; // can't do anything
  if (missing_q.size () > 0) 
    goto out; // don't find more missing keys, yet!
  if (replica_syncer && !replica_syncer->done())
    goto out; // replica syncer is still running.
  
  warn << "replica_maintenance_timer " << i << " is "
       << replicas [i]->id () << ", max " << ch_maxreps () << "\n";

  merkle_rep_tcb = NULL;
  rngmin = range_min[i];
  rngmax = host_node->my_ID ();
  
  blacklist::only().append_blacklist(host_node->my_location()->address(), exclu_proj);
  blacklist::only().append_blacklist(replicas[i]->address(), exclu_proj);
  specific_tree = New refcounted<merkle_tree> 
    (storage, rngmin, rngmax, exclu_proj);

  replica_syncer = New refcounted<merkle_syncer>
    (specific_tree, 
     wrap (this, &dhash_impl::doRPC_unbundler, replicas[i]),
     wrap (this, &dhash_impl::missing, replicas[i]));

  replica_syncer->sync (rngmin, rngmax);
  i++;

out:
  merkle_rep_tcb =
    delaycb (reptm (), wrap (this, &dhash_impl::replica_maintenance_timer, i));
}


dhash_fetchiter_res *
dhash_impl::block_to_res (dhash_stat err, s_dhash_fetch_arg *arg,
		          int cookie, ptr<dbrec> val)
{
  dhash_fetchiter_res *res;
  if (err) {
    res = New dhash_fetchiter_res  (DHASH_NOENT);
  } else {
    res = New dhash_fetchiter_res  (DHASH_COMPLETE);

    if (arg->start < 0)
      arg->start = 0;
    if (arg->start > val->len)
      arg->start = val->len;

    int n = (arg->len + arg->start < val->len) ? 
      arg->len : val->len - arg->start;

    res->compl_res->res.setsize (n);
    res->compl_res->attr.size = val->len;
    res->compl_res->offset = arg->start;
    res->compl_res->source = host_node->my_ID ();
    res->compl_res->cookie = cookie;
    memcpy (res->compl_res->res.base (), (char *)val->value + arg->start, n);

    //free the cookie if we just read the last byte
    pk_partial *part = pk_cache[cookie];
    if (part &&
	arg->len + arg->start == val->len) {
      pk_cache.remove (part);
      delete part;
    }

    bytes_served += n;
    keys_served++;
  }

  return res;
}

void
dhash_impl::fetchpk_cb (user_args *sbp, s_dhash_fetch_pk_arg *arg,
                        int cookie, ptr<dbrec> val, dhash_stat err)
{
  s_dhash_fetch_arg a;
  a.start = 0;
  a.len = arg->len;
  dhash_fetchiter_res *res = block_to_res (err, &a, cookie, val);
  sbp->reply (res);
  delete res;
}

void
dhash_impl::fetchiter_sbp_gotdata_cb (user_args *sbp, s_dhash_fetch_arg *arg,
				      int cookie, ptr<dbrec> val, 
				      dhash_stat err)
{
  dhash_fetchiter_res *res = block_to_res (err, arg, cookie, val);
  sbp->reply (res);
  delete res;
}

void
dhash_impl::dispatch (user_args *sbp) 
{
  rpc_answered++;
  switch (sbp->procno) {
  case DHASHPROC_OFFER:
    {
      dhash_offer_arg *arg = sbp->TEMPLATE getarg<dhash_offer_arg> ();
      dhash_offer_res res (DHASH_OK);
      res.resok->accepted.setsize (arg->keys.size ());
      for (u_int i = 0; i < arg->keys.size (); i++) {
	ref<dbrec> kkk = id2dbrec (arg->keys[i]);
	res.resok->accepted[i] = !storage->lookup_chblock (kkk);
	info << host_node->my_ID () << ": " << arg->keys[i]
	     << (res.resok->accepted[i] ? " not" : "") << " present\n";
      }

      sbp->reply (&res);
    }
    break;
  case DHASHPROC_FETCHITER:
    {
      s_dhash_fetch_arg *farg = sbp->TEMPLATE getarg<s_dhash_fetch_arg> ();

      if (farg->ctype == DHASH_KEYHASH) {
	bigint l;
      }

      blockID id (farg->key, farg->ctype, farg->dbtype);
      if ((key_status (id) != DHASH_NOTPRESENT) && (farg->len > 0))
        fetch (id, farg->cookie,
               wrap (this, &dhash_impl::fetchiter_sbp_gotdata_cb, sbp, farg));
      else {
        dhash_fetchiter_res res (DHASH_NOENT);
        sbp->reply (&res);
      }
    }
    break;

  case DHASHPROC_FETCH_PK:
    {
      s_dhash_fetch_pk_arg *arg =
	sbp->TEMPLATE getarg<s_dhash_fetch_pk_arg> ();
     
      ref<dbrec> k = id2dbrec(arg->key);
      ptr<dbrec> d = kidx->lookup (arg->key);
      long v0;
      
      if (kidx->getver (arg->key, v0)) {
	assert (d);
	if (v0 == (long)arg->vers) {
          keys_served++;
          dhash_fetchiter_res res (DHASH_SAMEVERSION);
          sbp->reply (&res);
	  return;
	}
	else if (v0 < (long)arg->vers) {
	  //warn << "fetch stale pk: " << v0 << " < " << (long)arg->vers << "\n";
          keys_served++;
          dhash_fetchiter_res res (DHASH_STALE);
          sbp->reply (&res);
	  return;
	}
      }
      else {
        dhash_fetchiter_res res (DHASH_NOENT);
        sbp->reply (&res);
	return;
      }

      blockID id (arg->key, DHASH_KEYHASH, DHASH_BLOCK);
      fetch (id, 0, wrap (this, &dhash_impl::fetchpk_cb, sbp, arg));
    }
    break;

  case DHASHPROC_CHECK_PKS:
    {
      s_dhash_check_pks_arg *arg =
	sbp->TEMPLATE getarg<s_dhash_check_pks_arg> ();

      if (blacklist::only().deny_project (host_node->my_location()->address(), arg->projkey)) {
	warn << "check_pks for unwilling host\n";
        dhash_fetchiter_res res (DHASH_UNWILLING);
        sbp->reply (&res);
	return;
      }

      if (arg->keys.size () != arg->versions.size ()) {
	warn << "size mismatch " << arg->keys.size ()
	     << " and " << arg->versions.size () << "\n";
	dhash_check_pks_res res (DHASH_ERR);
	sbp->reply (&res);
	return;
      }
          
      keys_served+=arg->keys.size ();

      vec<bigint> nv;
      vec<long> lv;
      vec<bigint> missing;

      for (unsigned i=0; i<arg->keys.size (); i++) {
	bigint n = arg->keys [i];
        long v = (long) arg->versions [i];
	long v0;
	if (kidx->getver (n, v0)) {
	  if (v0 != v) { // different block
	    nv.push_back (n);
	    lv.push_back (v0);
	  }
	}
	else
	  missing.push_back (n);
      }
      warn << "checked " << arg->keys.size () << " keys\n";

      dhash_check_pks_res res (DHASH_OK);
      res.resok->keys.setsize (nv.size ()+missing.size ());
      res.resok->versions.setsize (nv.size ());
      for (unsigned i=0; i<nv.size (); i++) {
	res.resok->keys [i] = nv [i];
	res.resok->versions [i] = lv [i];
      }
      for (unsigned i=0; i<missing.size (); i++)
	res.resok->keys [nv.size ()+i] = missing [i];
      sbp->reply (&res);
    }
    break;

  case DHASHPROC_OFFER_PKS:
    {
      s_dhash_offer_pks_arg *arg =
	sbp->TEMPLATE getarg<s_dhash_offer_pks_arg> ();

      if (arg->keys.size () != arg->versions.size ()) {
	warn << "size mismatch " << arg->keys.size ()
	     << " and " << arg->versions.size () << "\n";
	dhash_offer_pks_res res (DHASH_ERR);
	sbp->reply (&res);
	return;
      }

      vec<bigint> nv;
      for (unsigned i=0; i<arg->keys.size (); i++) {
	bigint n = arg->keys [i];
        long v = (long) arg->versions [i];
	long v0;
	if (kidx->getver (n, v0)) {
	  if (v0 < v) // older block - opposite of checkpks
	    nv.push_back (n);
	}
	else { // missing
	  nv.push_back (n);
	  warn << "offerpks: don't have " << n << "\n";
	}
      }

      dhash_offer_pks_res res (DHASH_OK);
      res.resok->keys.setsize (nv.size ());
      for (unsigned i=0; i<nv.size (); i++)
	res.resok->keys [i] = nv [i];
      sbp->reply (&res);
    }
    break;

  case DHASHPROC_STORE:
    {
      s_dhash_insertarg *sarg = sbp->TEMPLATE getarg<s_dhash_insertarg> ();

      if (sarg->type == DHASH_CACHE && 
	       ((sarg->ctype != DHASH_CONTENTHASH &&
		 sarg->ctype != DHASH_KEYHASH) ||
		sarg->dbtype != DHASH_BLOCK)) {
	dhash_storeres res (DHASH_ERR);
	ptr<location> pred = host_node->my_pred ();
	pred->fill_node (res.pred->p);
	sbp->reply (&res);
      } 
      else if ( blacklist::only().deny_project(host_node->my_location()->address(), sarg->projkey) ) {
	warn << "Refusing block " << sarg->key << " for project " << sarg->projkey << "\n";
	dhash_storeres res (DHASH_UNWILLING);
	sbp->reply (&res);
      }
      else {
        ref<dbrec> k = id2dbrec(sarg->key);
	bool exists =
	  dblookup (blockID (sarg->key, sarg->ctype, sarg->dbtype));
	store (sarg, exists,
	       wrap(this, &dhash_impl::storesvc_cb, sbp, sarg, exists));	
      }
    }
    break;
  case DHASHPROC_GETKEYS:
    {
      // XXX This is untested because it has been deprecated. (2-4-04)
      s_dhash_getkeys_arg *gkarg = sbp->TEMPLATE getarg<s_dhash_getkeys_arg>();
      dhash_getkeys_res res (DHASH_OK);

      // get all keys btw start and pred_id
      vec < ptr<dbrec> > keylist;
      storage->getchblockids(&keylist, id2dbrec(gkarg->start), id2dbrec(gkarg->pred_id), 
			     1024 / sha1::hashsize);

      ref<vec<chordID> > keys = New refcounted<vec<chordID> >;
      unsigned int i;
      for(i=0; i<keylist.size(); i++) {
	warn << "get keys " << i << " returns " << dbrec2id(keylist[i]) << "\n";
	keys->push_back( dbrec2id(keylist[i]) );
      }

      res.resok->keys.set (keys->base (), keys->size ());
      sbp->reply (&res);
    }
    break;
  default:
    sbp->replyref (PROC_UNAVAIL);
    break;
  }
}

void
dhash_impl::storesvc_cb (user_args *sbp,
		         s_dhash_insertarg *arg,
		         bool already_present,
		         dhash_stat err)
{
  dhash_storeres res (DHASH_OK);
  if ((err != DHASH_OK) && (err != DHASH_STORE_PARTIAL)) 
    res.set_status (err);
  else {
    res.resok->already_present = already_present;
    res.resok->source = host_node->my_ID ();
    res.resok->done = (err == DHASH_OK);
  }
  sbp->reply (&res);
}


//---------------- no sbp's below this line --------------
 
// -------- reliability stuff

// --- node to database transfers --- 

void
dhash_impl::fetch (blockID id, int cookie, cbvalue cb) 
{
  // if the cookie is in the hash, return that value
  pk_partial *part = pk_cache[cookie];
  if (part) {
    cb (cookie, part->val, DHASH_OK);
    return;
  }

  ptr<dbrec> ret = dblookup (id);
  if (!ret) {
    (*cb)(cookie, NULL, DHASH_NOENT);
  }
  
  else {
    // make up a cookie and insert in hash if this is the first fetch
    // of a KEYHASH
    if ((cookie == 0) && 
	id.ctype == DHASH_KEYHASH) {
      pk_partial *part = New pk_partial (ret, pk_partial_cookie);
      pk_partial_cookie++;
      pk_cache.insert (part);
      (*cb)(part->cookie, ret, DHASH_OK);
    }
    else
      (*cb)(-1, ret, DHASH_OK);
    
  }
}

void
dhash_impl::append (ref<dbrec> key, ptr<dbrec> data,
		    s_dhash_insertarg *arg,
		    cbstore cb)
{
  blockID id (arg->key, arg->ctype, arg->dbtype);

  if (key_status (id) == DHASH_NOTPRESENT) {
    // create a new record in the database
    xdrsuio x;
    char *m_buf;
    if ((m_buf = (char *)XDR_INLINE (&x, data->len + 3 & ~3))) {
      memcpy (m_buf, data->value, data->len);
      int m_len = x.uio ()->resid ();
      char *m_dat = suio_flatten (x.uio ());
      ref<dbrec> marshalled_data = New refcounted<dbrec> (m_dat, m_len);

      keys_others += 1;

      // XXX need to verify projkey is not on blacklist. 
      int ret = db_insert_immutable (arg->projkey, key, 
				     marshalled_data, DHASH_APPEND);
      assert (!ret);
      append_after_db_store (cb, arg->key, 0);
      xfree (m_dat);
    }
    else {
      warn << "DHASH_STOREERR 1\n";
      cb (DHASH_STOREERR);
    }
  }
  else
    fetch (id, -1, wrap (this, &dhash_impl::append_after_db_fetch,
			 key, data, arg, cb));
}

void
dhash_impl::append_after_db_fetch (ref<dbrec> key, ptr<dbrec> new_data,
			           s_dhash_insertarg *arg, cbstore cb,
			           int cookie, ptr<dbrec> data, dhash_stat err)
{
  if (arg->ctype != DHASH_APPEND)
    cb (DHASH_STORE_NOVERIFY);
  
  else {
    ptr<dhash_block> b = get_block_contents (data, DHASH_APPEND);
    xdrsuio x;
    char *m_buf;
    if ((data->len+b->len <= 64000) &&
	(m_buf = (char *)XDR_INLINE (&x, data->len+b->len))) {
      memcpy (m_buf, b->data, b->len);
      memcpy (m_buf + b->len, new_data->value, new_data->len);
      int m_len = x.uio ()->resid ();
      char *m_dat = suio_flatten (x.uio ());
      ptr<dbrec> marshalled_data =
        New refcounted<dbrec> (m_dat, m_len);

      // XXX need to verify projkey is not on blacklist. 
      int ret = db_insert_immutable (arg->projkey, key, 
				     marshalled_data, DHASH_APPEND);
      assert (!ret);
      append_after_db_store (cb, arg->key, 0);
      xfree (m_dat);
    }
    else {
      warn << "DHASH_STOREERR 2\n";
      cb (DHASH_STOREERR);
    }
  }

}

void
dhash_impl::append_after_db_store (cbstore cb, chordID k, int stat)
{
  if (stat){
    warn << "DHASH_STOREERR 3\n";
    cb (DHASH_STOREERR);
  } else
    cb (DHASH_OK);

  store_state *ss = pst[k];
  if (ss) {
    pst.remove (ss);
    delete ss;
  }
  //replicate?
}


void 
dhash_impl::store (s_dhash_insertarg *arg, bool exists, cbstore cb)
{
  if (arg->ctype == DHASH_CONTENTHASH && exists) {
    cb (DHASH_OK);
    return;
  }

  store_state *ss = pst[arg->key];
 
  if (ss == NULL) {
    ss = New store_state (arg->key, arg->attr.size);
    pst.insert(ss);
  }

  if (!ss->addchunk(arg->offset, arg->offset+arg->data.size (), 
		    arg->data.base ())) {
    cb (DHASH_ERR);
    return;
  }

  if (ss->iscomplete()) {

    dhash_stat stat = DHASH_OK;
    ref<dbrec> k = id2dbrec(arg->key);
    ref<dbrec> d = New refcounted<dbrec> (ss->buf, ss->size);

    switch (arg->ctype) {
    case DHASH_KEYHASH:
      {
	if (!verify_keyhash (arg->key, ss->buf, ss->size)) {
	  warning << "keyhash: cannot verify " << ss->size << " bytes\n";
	  stat = DHASH_STORE_NOVERIFY;
	  break;
	}

	stat = kidx->insert (arg->projkey, k, d);
	if (arg->type != DHASH_STORE)
	  stat = DHASH_OK;
	else if (stat == DHASH_OK)
	  info << "db write: " << host_node->my_ID ()
	       << " U " << arg->key << " " << ss->size << "\n";
	break;
      }

    case DHASH_APPEND:
      pst.remove (ss);
      delete ss;
      append (k, d, arg, cb);
      return;

    case DHASH_CONTENTHASH:
    case DHASH_NOAUTH:
    case DHASH_UNKNOWN:
      store_2 (arg, ss, k, d, stat, cb);
      return;

    default:
      stat = DHASH_ERR;
      break;
    }

    store_3 (arg, ss, stat, cb);
    return;

  }
  else
    cb (DHASH_STORE_PARTIAL);
}

void
dhash_impl::check_prblock_cb (ptr<prblock_args> args, dhash_stat stat, 
			      ptr<dhash_block> b, route r) {
  // read aqua's key (XXX this needs to be moved somewhere sane)
  str rawkey = file2str ("aqua_sk");
  ptr<sfspriv> aqua_sk = sfscrypt.alloc_priv (rawkey, SFS_SIGN);
  ptr<sfs_pubkey2> aqua_pk = New refcounted<sfs_pubkey2>();
  aqua_sk->export_pubkey(aqua_pk);
  ptr<sfspub> verifier = sfscrypt.alloc(*aqua_pk, SFS_VERIFY);

  // If we can't retrieve the project block, reject.
  if (stat != DHASH_OK) {
    warn << "problem retrieving project block: 0x" << args->id << "\n";
    args->stat =   DHASH_STORE_NOVERIFY;
    store_3 (args->arg, args->ss, args->stat, args->cb);
    return;
  }

  // if writer_pk is not in project block, then stat = DHASH_STORE_NOVERIFY;
  if (b->ctype != DHASH_KEYHASH) {
    store_3 (args->arg, args->ss, DHASH_STORE_NOVERIFY, args->cb);
    return;
  }

  // verify that the project pk was signed by Aqua admin
  strbuf pkstrbuf;
  ptr<sfs_pubkey2> project_pk = keyhash_pk(b->data, b->len);
  ptr<sfspub> project_general_pk = sfscrypt.alloc (*project_pk, SFS_VERIFY);
  project_general_pk->export_pubkey(pkstrbuf, false);
  str projectpkstr(pkstrbuf);
  
  ptr<dhash_block> contents = get_block_contents (b, DHASH_KEYHASH);
  sig_hashlist parsed;
  xdrmem x2(contents->data, contents->len);

  ptr<keyhash_payload> payload = keyhash_payload::decode(contents);
  xdrmem x( payload->buf().cstr(), payload->buf().len() );

  if (!parsed.decode(x) ||
      !verifier->verify(parsed.sig, projectpkstr) ||
      !parsed.contains
        (compute_hash (args->writer_pk.cstr(), args->writer_pk.len()))) {
    store_3 (args->arg, args->ss, DHASH_STORE_NOVERIFY, args->cb);
    return;
  }
  
  // else, store the block
  warn << "retrieved the project block. storing the new block.\n";
  store_2 (args->arg, args->ss, args->k, args->d, args->stat, args->cb);
}

void
dhash_impl::store_2 (s_dhash_insertarg *arg, store_state *ss, 
		     ptr<dbrec> k, ptr<dbrec> d,
		     dhash_stat stat, cbstore cb)
{
  // Actually stores the block
  int ret = db_insert_immutable (arg->projkey, k, d, arg->ctype);
  if (ret != 0) {
    warning << "db write failure: " << db_strerror (ret) << "\n";
    warn << "DHASH_STOREERR 4\n";
    stat = DHASH_STOREERR;
  }

  store_3 (arg, ss, stat, cb);
  return;
}

void 
dhash_impl::store_3 (s_dhash_insertarg *arg, store_state *ss, 
		     dhash_stat stat, cbstore cb)
{
  // Reply to the requestor
  if (stat == DHASH_OK) {
    if (arg->type == DHASH_STORE ||
	arg->type == DHASH_FRAGMENT) {
      keys_stored ++;
      bytes_stored += ss->size;
    }
    else
      keys_others ++;
  }

  pst.remove (ss);
  delete ss;

  cb (stat);
}

// --------- utility

dhash_stat
dhash_impl::key_status (const blockID &n)
{
  ptr<dbrec> val = dblookup (n);
  if (!val)
    return DHASH_NOTPRESENT;
  return DHASH_OK;
}

void
dhash_impl::doRPC_unbundler (ptr<location> dst, RPC_delay_args *args)
{
  host_node->doRPC
    (dst, args->prog, args->procno, args->in, args->out, args->cb);
}


void
dhash_impl::doRPC (const chord_node &n, const rpc_program &prog, int procno,
	           ptr<void> in,void *out, aclnt_cb cb,
		   cbtmo_t cb_tmo = NULL) 
{
  host_node->doRPC (n, prog, procno, in, out, cb, cb_tmo);
}

void
dhash_impl::doRPC (const chord_node_wire &n, const rpc_program &prog,
                   int procno, ptr<void> in,void *out, aclnt_cb cb,
		   cbtmo_t cb_tmo = NULL) 
{
  host_node->doRPC (make_chord_node (n), prog, procno, in, out, cb, cb_tmo);
}

void
dhash_impl::doRPC (ptr<location> ID, const rpc_program &prog, int procno,
	           ptr<void> in,void *out, aclnt_cb cb,
		   cbtmo_t cb_tmo = NULL)  
{
  host_node->doRPC (ID, prog, procno, in, out, cb, cb_tmo);
}


// ---------- debug ----

void
dhash_impl::get_stats (strbuf& x)
{
#if 0
  warnx << "ID: " << host_node->my_ID () << "\n";
  warnx << "Stats:\n";
  warnx << "  " << keys_stored << " blocks stored\n";
  warnx << "  " << keys_others << " other stored\n";
  warnx << "  " << bytes_stored << " total bytes held\n";
  warnx << "  " << keys_served << " keys served\n";
  warnx << "  " << bytes_served << " bytes served\n";
  warnx << "  " << rpc_answered << " rpc answered\n";
#endif

  x << "host:" << host_node->my_ID () << "\n";
  x << "keys stored:" << keys_stored << "\n";
  x << "bytes stored:" << bytes_stored << "\n";
  x << "keys served:" << keys_served << "\n";
  x << "bytes served:" << bytes_served << "\n";
  keys_stored = 0;
  bytes_stored = 0;
  keys_served = 0;
  bytes_served = 0;
}

void
dhash_impl::stop ()
{
  if (keyhash_mgr_tcb) {
    warnx << "stop replica timer\n";
    timecb_remove (keyhash_mgr_tcb);
    keyhash_mgr_tcb = NULL;
  }
  if (merkle_rep_tcb) {
    timecb_remove (merkle_rep_tcb);
    merkle_rep_tcb = NULL;
    if (pmaint_obj)
      pmaint_obj->stop ();
    warn << "stop merkle replication timer\n";
  }
}

ptr<dbrec>
dhash_impl::dblookup (const blockID &i)
{
  if(i.ctype == DHASH_KEYHASH)
    return kidx->lookup (i.ID);
  else
    return storage->lookup_chblock (id2dbrec (i.ID));
}

int
dhash_impl::db_insert_immutable (bigint projkey, ref<dbrec> key, 
				 ref<dbrec> data, dhash_ctype ctype)
{
  assert(!blacklist::only().deny_project(host_node->my_location()->address(), projkey));
  char *action;
  block blk (to_merkle_hash (key), data);

  int ret = 0;
  if (storage->lookup_chblock (key, NULL)) {
    action = "R";
  } else {
    action = "N"; // New
    ret = storage->insert_chblock(projkey, key, data);
  }

  // Won't deal well if there's a magic expansion in the encoding
  // vector.  Should really know how big the full block is so that
  // the right amount of the key's encoding vector is extracted...
  str x ("");
  if (key->isFrag() &&
      (u_long) data->len > 9 + 2 * num_dfrags ())
    x = strbuf () << " " << hexdump (data->value + 8, 2*(num_dfrags () + 1));
  info << "db write: " << host_node->my_ID ()
       << " " << action << " " << dbrec2id(key) << x << "\n";
  return ret;
}

void
dhash_impl::db_delete_immutable (ref<dbrec> key)
{
  assert (storage->lookup_chblock (key, NULL));
  storage->remove_chblock (key);

  info << "db delete: " << host_node->my_ID ()
       << " " << dbrec2id(key) << "\n";
}

// ----------------------------------------------------------------------------
// store state 

static void
join (store_chunk *c)
{
  store_chunk *cnext;

  while (c->next && c->end >= c->next->start) {
    cnext = c->next;
    if (c->end < cnext->end)
      c->end = cnext->end;
    c->next = cnext->next;
    delete cnext;
  }
}

bool
store_state::gap ()
{
  if (!have)
    return true;

  store_chunk *c = have;
  store_chunk *p = 0;

  if (c->start != 0)
    return true;

  while (c) {
    if (p && p->end != c->start)
      return true;
    p = c;
    c = c->next;
  }
  if (p->end != size)
    return true;
  return false;
}

bool
store_state::iscomplete ()
{
  return have && have->start == 0 && have->end == (unsigned)size && !gap ();
}

bool
store_state::addchunk (unsigned int start, unsigned int end, void *base)
{
  store_chunk **l, *c;

  if (start >= end || end > size)
    return false;
  
  l = &have;
  for (l=&have; *l; l=&(*l)->next) {
    c = *l;
    // our start touches this block
    if (c->start <= start && start <= c->end) {
      // we have new data
      if (end > c->end) {
        memmove (buf+start, base, end-start);
        c->end = end;
        join(c);
      }
      return true;
    }
    // our start comes before this block; break to insert
    if (start < c->start)
      break;
  }
  *l = New store_chunk(start, end, *l);
  memmove(buf+start, base, end-start);
  join(*l);
  return true;
}

