/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "async.h"
#include "str.h"
#include "pastdbfe.h"
#include "db.h"
#include <db_wrap.h>

#define CACHE_OPT  "opt_cachesize"
#define PERM_OPT   "opt_permissions"
#define TYPE_OPT   "opt_dbtype"
#define CREATE_OPT "opt_create"
#define FLAG_OPT   "opt_flag"
#define JOIN_OPT   "opt_join" 

/*
dbOptions::dbOptions() {}

int
dbOptions::addOption(char *optionSig, long value)
{
  optionRec optr;
  optr.sig = optionSig;
  optr.value = value;
  options.push_back(optr);
  return 0;
}

long
dbOptions::getOption(char *optionSig)
{
  for (unsigned int i=0; i < options.size(); i++) 
    if (memcmp(options[i].sig, optionSig, strlen(options[i].sig)) == 0) 
      return options[i].value;
  return -1;
}


//////////////////////pdbEnumeration//////////////////////
*/

pdbEnumeration::pdbEnumeration(DB *db, DB_ENV *dbe)
{
  db_sync = db;

  int r = db->cursor(db, NULL, &cursor, 0);
  if (r && dbe) { 
    dbe->err(dbe, r, "db->cursor");
  }
  cursor_init = 0;
}

pdbEnumeration::~pdbEnumeration()
{
  cursor->c_close (cursor);
}

ptr<dbrec>
pdbEnumeration::getElement(u_int32_t flags, ptr<dbrec> startkey)
{
  DBT key, data;
  bzero(&key, sizeof(key));
  if (startkey) {
    key.size = startkey->len;
    key.data = startkey->value;
  }
  bzero(&data, sizeof(data));
  data.flags = DB_DBT_PARTIAL;
  int err = cursor->c_get(cursor, &key, &data, flags);
  cursor_init = 1;
  if (err)
    return NULL;
  ref<dbrec> keyrec = New refcounted<dbrec>(key.data, key.size);
  return keyrec;
}

ptr<dbrec>
pdbEnumeration::nextElement()
{
  return getElement (DB_NEXT, NULL);
}

ptr<dbrec>
pdbEnumeration::prevElement()
{
  return getElement (DB_PREV, NULL);
}

ptr<dbrec>
pdbEnumeration::nextElement(ref<dbrec> startkey)
{
  return getElement(DB_SET_RANGE, startkey);
}

ptr<dbrec>
pdbEnumeration::lastElement()
{
  return getElement(DB_LAST, NULL);
}

ptr<dbrec>
pdbEnumeration::firstElement()
{
  return getElement(DB_FIRST, NULL);
}


////////////////////// pastdbfe //////////////////////////////

pastdbfe::pastdbfe()
{
  db = NULL;
  dbe = NULL;
  _tx = false;
}

pastdbfe::~pastdbfe()
{
  closedb ();
}

int 
pastdbfe::opendb_env (const char *filename, dbOptions opts)
{
  char cpath[MAXPATHLEN];
  int r = -1, dfd = -1;

  long mode = opts.getOption(PERM_OPT);
  if (mode == -1) mode = 0664;
  long flags = opts.getOption(FLAG_OPT);
  if (flags == -1) flags = DB_CREATE;
  long create = opts.getOption(CREATE_OPT);
  if (create == 0) flags |= DB_EXCL;
  long cacheSize = opts.getOption(CACHE_OPT);
  long join = opts.getOption(JOIN_OPT);

  dfd = open (".", O_RDONLY);
  if (dfd == -1)
    fatal << "can't open current working dir\n";
  mkdir (filename, 0755);

  if (filename) {
    lockf = str (filename, strlen (filename)) << ".lck";
    lockfd = open (lockf, O_CREAT | O_TRUNC | O_EXCL, 0644);
    if (lockfd < 0) {
      warn << "cannot open lock file " << lockf << ", db in use?\n";
      return -1;
    }
  }
  else
    lockfd = -1;

  r = chdir (filename);
  if(r == -1)
    fatal << "couldn't chdir to " << filename << "\n";

  r = db_env_create (&dbe, 0);
  if (r) {
    warn << "cannot create env\n";
    goto done;
  }

  if (cacheSize > -1) {
    r = dbe->set_cachesize (dbe, 0, cacheSize*1024, 0);
    if (r) {
      warn << "cannot set cache size\n";
      goto done;
    }
  }

  getcwd (cpath, MAXPATHLEN);
  r = dbe->set_data_dir (dbe, cpath);
  r = dbe->set_lg_dir(dbe, cpath);
  r = dbe->set_tmp_dir(dbe, cpath); 
  if (r) {
    warn << "cannot set data dir\n";
    goto done;
  }

  if (join < 0)
    r = dbe->open
      (dbe, NULL, DB_CREATE | DB_INIT_MPOOL | DB_INIT_LOCK |
                  DB_INIT_LOG | DB_INIT_TXN | DB_RECOVER, 0);
  else
    r = dbe->open (dbe, NULL, DB_JOINENV, 0);

  if (r) {
    warn << "cannot open env\n";
    goto done;
  }

  r = db_create (&db, dbe, 0);
  if (r) {
    warn << "cannot create db\n";
    goto done;
  }

  _tx = false;

  db->set_pagesize(db, 20 * 1024);
  r = open_database (db, NULL, "db", NULL, DB_BTREE, flags, mode);
  
  if (r)
    warn << "cannot open db\n";

done:
  if (r)
    warn << "db error: " << db_strerror(r) << "\n";
  fchdir (dfd);
  close (dfd);
  return r;
}

int 
pastdbfe::opendb (dbOptions opts)
{
  return opendb (NULL, opts);
}

int 
pastdbfe::opendb (const char *filename, dbOptions opts)
{ 
  int r = -1;

  long mode = opts.getOption(PERM_OPT);
  if (mode == -1) mode = 0664;
  long flags = opts.getOption(FLAG_OPT);
  if (flags == -1) flags = DB_CREATE;
  long create = opts.getOption(CREATE_OPT);
  if (create == 0) flags |= DB_EXCL;
  long cacheSize = opts.getOption(CACHE_OPT);

  r = db_create (&db, dbe, 0);
  if (r) return r;

  db->set_pagesize (db, 20 * 1024);

  if (cacheSize > -1) {
    r = db->set_cachesize(db, 0, cacheSize*1024, 0);
    if (r) return r;
  }

  if (filename) {
    lockf = str (filename, strlen (filename)) << ".lck";
    lockfd = open (lockf, O_CREAT | O_TRUNC | O_EXCL, 0644);
    if (lockfd < 0) {
      warn << "cannot open lock file " << lockf << ", db in use?\n";
      return -1;
    }
  }
  else
    lockfd = -1;
  r = open_database (db, NULL, filename, NULL, DB_BTREE, flags, mode);
  return r;
}

int
pastdbfe::insert (ref<dbrec> key, ref<dbrec> data)
{
  DBT skey, content;
  bzero(&skey, sizeof(skey));
  bzero(&content, sizeof(content));
  content.size = data->len;
  content.data = data->value;
  skey.size = key->len;
  skey.data = key->value;
  
  DB_TXN *t = NULL;
  int r = 0;
  int tr = 0;

  if (_tx) {
    r = begin_transaction(dbe, NULL, &t, 0);
    if (r) return r;
  }

  r = db->put (db, t, &skey, &content, 0);

  if (t) {
    tr = commit_transaction(t, 0);
    if (!r) r = tr;
    t = NULL;
  }

  return r;
} 

ptr<dbrec> 
pastdbfe::lookup (ref<dbrec> key) 
{ 
  DBT skey, content;
  bzero(&skey, sizeof(skey));
  skey.size = key->len;
  skey.data = key->value;
  bzero(&content, sizeof(content));
  int r=0;

  if ((r = db->get (db, NULL, &skey, &content, 0)) != 0)
    return NULL;
  
  ptr<dbrec> ret = New refcounted<dbrec>(content.data, content.size);
  return ret;
} 

int 
pastdbfe::remove (ref<dbrec> key)
{
  DBT dkey;
  bzero(&dkey, sizeof(dkey));
  dkey.size = key->len;
  dkey.data = key->value;

  DB_TXN *t = NULL;
  int r = 0;
  int tr = 0;

  if (_tx) {
    r = begin_transaction( dbe, NULL, &t, 0);
    if (r) return r;
  }

  r = db->del (db, NULL, &dkey, 0);

  if (t) {
    tr = commit_transaction(t, 0);
    if (!r) r = tr;
    t = NULL;
  }

  return r;
}

int 
pastdbfe::closedb ()
{
  if (lockfd > -1) {
    close (lockfd);
    unlink (lockf);
    lockfd = -1;
  }
  int r = 0;
  if (db) {
    r = db->close(db, 0);
    db = 0;
  }
  if (dbe) {
    dbe->close(dbe, 0);
    dbe = 0;
  }
  return r;
}

ptr<pdbEnumeration> 
pastdbfe::enumerate ()
{
#if 0
  if (_tx)
    return 0L;
#endif
  return New refcounted<pdbEnumeration>(db, dbe);
}

void
pastdbfe::sync ()
{
  db->sync (db, 0L);
}

