/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include "pastwatch.h"
#include "io.h"

typedef callback<void, bigint, int>::ref fhcb_t;
typedef callback<void, const pde_t &, fhcb_t >::ref recursecb_t;

extern void recurse (str dirpath, recursecb_t iter_cb, cbi cb, bigint dirfh, 
		     unsigned max_outstanding);
extern mtime_t write_file
  (str, bigint, bigint, callback<void, cfsrw_status>::ref);

extern vec<str> notify_cmds;
extern vec<str> notify_files;

static void
write_file_done (str path, time_t mt, fhcb_t fhcb, bigint fh,
                 cfsrw_status err)
{
  if (err) {
    warn << "cannot write file into repository\n";
    fhcb (fh, -1);
    return;
  }
  
  pastwatch_status ("N", path);
  fhcb (fh, 0);
}

static void
create_reply (str path, time_t mt, fhcb_t fhcb,
              cfswcallp c, cfsrw_status status)
{
  extern bigint arg_view;
  if (status) {
    bigint blah;
    fhcb (blah, -1);
    return;
  }

  cfsw_create_res *res = static_cast<cfsw_create_res *> (c->res ());
  write_file (path, res->object, arg_view,
              wrap (write_file_done, path, mt, fhcb, res->object));
}

static void
mkdir_reply (str path, time_t mt, fhcb_t fhcb,
             cfswcallp c, cfsrw_status status)
{
  if (status) {
    bigint blah;
    fhcb (blah, -1);
    return;
  }
 
  cfsw_create_res *res = static_cast<cfsw_create_res *> (c->res ());
  pastwatch_status ("N", path);
  fhcb (res->object, 0);
}

static void
symlink_reply (str path, time_t mt, fhcb_t fhcb,
               cfswcallp c, cfsrw_status status)
{
  if (status) {
    bigint blah;
    fhcb (blah, -1);
    return;
  }
  
  cfsw_symlink_res *res = static_cast<cfsw_symlink_res *> (c->res ());
  pastwatch_status ("N", path);
  fhcb (res->object, 0);
}

static void
check_objid (const pde_t& e, fhcb_t fhcb, int tries);

static void
checkin_getattr_reply (const pde_t e, fhcb_t fhcb, bigint newid, int tries,
                       cfsrcallp cc, cfsrw_status status)
{
  if (status != CFSRWERR_NOENT) {
    if (tries == 5) {
      warn << "cannot compute an object handle that has not been used before\n";
      bigint blah;
      fhcb (blah, EIO);
    }
    else 
      check_objid (e, fhcb, tries+1);
    return;
  }

  struct stat sb;
  
  if (int err = lstat (e.path, &sb)) {
    //warn << "checking: " << e.path << "\n";
    perror ("lstat");
    bigint blah;
    fhcb (blah, err);
    return;
  }

  switch (e.type) {
    case CFSDIR:
      {
	cfswcallp c = New refcounted<cfswcall>
	  (CFSWPROC_MKDIR, wrap (mkdir_reply, e.path, sb.st_mtime, fhcb));
	cfsw_create_arg *arg = static_cast<cfsw_create_arg *> (c->arg ());
	arg->dir = e.dirfh;
	arg->name = e.name;
	arg->object = newid;
	arg->attributes.valid_size = 1;
	arg->attributes.size = sb.st_size;
	arg->attributes.valid_mode = 1;
	arg->attributes.mode = sb.st_mode;
	cfsw_call (c);
      }
      break;
    case CFSREG:
      {
	cfswcallp c = New refcounted<cfswcall>
	  (CFSWPROC_CREATE, wrap (create_reply, e.path, sb.st_mtime, fhcb));
	cfsw_create_arg *arg = static_cast<cfsw_create_arg *> (c->arg ());
	arg->dir = e.dirfh;
	arg->name = e.name;
	arg->object = newid;
	arg->attributes.valid_size = 1;
	arg->attributes.size = sb.st_size;
	arg->attributes.valid_mode = 1;
	arg->attributes.mode = sb.st_mode;
	cfsw_call (c);
      }
      break;
    case CFSLNK:
      {
	char buf[256];
	int r;
	if ((r = readlink (e.path, buf, 256)) < 0) {
          perror ("readlink");
	  bigint blah;
          fhcb (blah, r);
	  return;
	}
	str link (buf, r);

	cfswcallp c = New refcounted<cfswcall>
	  (CFSWPROC_SYMLINK, wrap (symlink_reply, e.path, sb.st_mtime, fhcb));
	cfsw_symlink_arg *arg = static_cast<cfsw_symlink_arg *> (c->arg ());
	arg->dir = e.dirfh;
	arg->name = e.name;
	arg->linkto = link;
	arg->object = newid;
	arg->attributes.valid_size = 1;
	arg->attributes.size = sb.st_size;
	arg->attributes.valid_mode = 1;
	arg->attributes.mode = sb.st_mode;
	cfsw_call (c);
      }
      break;
    default:
      assert (0);
  }
}

static void
check_objid (const pde_t& e, fhcb_t fhcb, int tries)
{
  // compute random handle
  bigint n;
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&n, xxb, 20);
  
  cfsrcallp c = New refcounted<cfsrcall>
    (CFSRPROC_GETATTR, wrap (checkin_getattr_reply, e, fhcb, n, tries));
  cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
  arg->object = n;
  cfsr_call (c);
}

void
checkin (const pde_t& e, fhcb_t fhcb)
{
  //warn << "checkin called " << ++count << " times\n";
  check_objid (e, fhcb, 0);
}

static void
link_reply (bigint fh, str path, fhcb_t fhcb,
            cfswcallp c, cfsrw_status status)
{
  if (status) {
    bigint blah;
    fhcb (blah, -1);
    return;
  }
  fhcb (fh, 0);
}

void
relink (const pde_t& e, fhcb_t fhcb)
{
  assert (e.type == CFSREG);
  cfswcallp c = New refcounted<cfswcall>
    (CFSWPROC_LINK, wrap (link_reply, e.fh, e.path, fhcb));
     cfsw_link_arg *arg = static_cast<cfsw_link_arg *> (c->arg ());
  arg->dir = e.dirfh;
  arg->name = e.name;
  arg->object = e.fh;
  cfsw_call (c);
}

static void
mkroot_reply (bigint root, str path, cbi cb,
              cfswcallp c, cfsrw_status status)
{
  if (status) {
    warn << "cannot create " << path << " in repository.\n";
    cb (-1);
    return;
  }
  // for commit notification message
  notify_cmds.push_back ("A");
  notify_files.push_back (path);

  cfsw_create_res *res = static_cast<cfsw_create_res *> (c->res ());
  recurse (".", wrap (checkin), cb, res->object, 200);
  //warn << "recurse returned\n";
}

static void
lookup_reply (bigint root, str path, cbi cb,
              cfsrcallp c, cfsrw_status status)
{
  if (!status || status != CFSRWERR_NOENT) {
    if (!status)
      warn << path << " already exists in repository.\n";
    else
      warn << "cannot lookup " << path << " in repository.\n";
    cb (-1);
    return;
  }

  // status == CFSRWERR_NOENT
  cfsw_mkdir (root, path, wrap (mkroot_reply, root, path, cb));
}

void
import (bigint root, str path, cbi cb)
{
  if (strchr(path.cstr(), '/')) {
    warn << path << ": bad module name, cannot specify a subdirectory.\n";
    cb (-1);
    return;
  }
  
  // cfsr_lookup breaks <path> into individual directories, and
  // performs a lookup on each one.
  cfsr_lookup (root, path, wrap (lookup_reply, root, path, cb));
}

