/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

/* this file calls the GNU diff and diff3 utilities. */

#include <string.h>
#include <stdio.h>
#include "diffrun.h"

#define DIFF  "diff"
#define DIFF3 "diff3"

#include "path.h"

//
// wrappers for calling diff
//

static void
call_diff_write_stdout (const char *text)
{
  warn << text << "\n";
}

static void
call_diff_error (const char *format, const char *a1, const char *a2)
{
  warn (format, a1, a2);
}

static struct diff_callbacks call_diff_file_callbacks =
{
#if 0
  (void (*) PROTO((const char *, size_t))) NULL,
  (void (*) PROTO((void))) NULL,
#else
  NULL, NULL,
#endif
  call_diff_write_stdout,
  call_diff_error
};


//
// diff calls for pastwatch code
//

static int
diff (int argc, const char **argv, strbuf &b)
{
  str s = pastwatch_tmpfile ();
  int error =
    diff_run (argc, (char**) argv, (char*) s.cstr (),
	      &call_diff_file_callbacks);

  char buf [512];
  if (error >= 0) {
    int fd = open (s, O_RDWR, 0600);
    if (fd < 0) {
      warn << "cannot open " << s << " for reading\n";
      error = -1;
    }
    else {
      int r;
      while ((r = read (fd, buf, 512)) > 0) {
	str x (buf, r);
	b << x;
      }
      if (r < 0) {
	warn << "read/write error while merging\n";
	error = -1;
      }
    }
    close (fd);
  }

  unlink (s);
  return error;
}

int
diff_context
  (const char *f1, const char *f2, strbuf &b, str t1)
{
  if (t1) {
    int argc = 6;
    const char *argv[6] = { "diff", "-c", "-L", t1.cstr (), f1, f2 };
    return diff (argc, argv, b);
  }
  else {
    int argc = 4;
    const char *argv[4] = { "diff", "-c", f1, f2 };
    return diff (argc, argv, b);
  }
}

int
diff_normal (const char *f1, const char *f2, strbuf &b)
{
  int argc = 3;
  const char *argv[3] = { "diff", f1, f2 };
  return diff (argc, argv, b);
}

int
isdiff (const char *f1, const char *f2)
{
  int argc = 4;
  const char *argv[4] = { "diff", "-q", f1, f2 };
  strbuf b;
  int r = diff (argc, argv, b);
  if (r != -1) {
    str x = b;
    r = x.len ();
  }
  return r;
}

int
diff (const char *f1, const char *f2, strbuf& b)
{
  int argc = 3;
  const char *argv[3] = { "diff", f1, f2 };
  return diff (argc, argv, b);
}


//
// merge () uses diff3 to merge two files from an older version.
// returns 0 if successful, -1 if error, 1 if there are conflicts.
//

int
merge (const char *fb, const char *tfb,
       const char *fw, const char *tfw,
       const char *fr, const char *tfr,
       bool check_only)
{
  int argc = 13;
  const char *argv[13] = {
    "diff3", "-E", "-a", "-m",
    "-L", tfw,
    "-L", tfb,
    "-L", tfr,
    fw, fb, fr
  };

  str s = pastwatch_tmpfile ();
  int conflict =
    diff3_run (argc, (char**) argv, (char*) s.cstr (),
	       &call_diff_file_callbacks);

  char buf[512];
  int error = -1;

  if (conflict == 0)
    error = 0;
  else if (conflict > 0)
    error = 1;
  else
    error = -1;

  if (error >= 0 && !check_only) {
    int fd = open (s, O_RDWR, 0600);
    if (fd < 0) {
      warn << "cannot open " << s << " for reading\n";
      error = -1;
    }
    int wfd = open (fw, O_TRUNC | O_WRONLY);
    if (wfd >= 0 && fd >= 0) {
      int r;
      while ((r = read (fd, buf, 512)) > 0) {
	if ((r = write (wfd, buf, r)) < 0)
	  break;
      }
      if (r < 0) {
	warn << "read/write error while merging\n";
	error = -1;
      }
      close (wfd);
    }
    else if (wfd < 0) {
      warn << "cannot open " << fw << " for writing\n";
      error = -1;
    }
    close (fd);
  }

  unlink (s);
  return error;
}

/* code stolen from cvs, modified by benjie */

/*
 * Copyright (c) 1992, Brian Berliner and Jeff Polk
 * Copyright (c) 1989-1992, Brian Berliner
 * 
 * You may distribute under the terms of the GNU General Public License as
 * specified in the README file that comes with the CVS source distribution.
 * 
 * Various useful functions for the CVS support code.
 */

#define DIFF3_MERGE_PAT_1 "<<<<<<< "
#define DIFF3_MERGE_PAT_2 "=======\n"
#define DIFF3_MERGE_PAT_3 ">>>>>>> "

int
isconf (const char *path)
{
  FILE *fp;
  char line [2048];
  int result = 0;

  fp = fopen (path, "r");
  if (fp == NULL)
    return -1;

  while (fgets (line, 2048, fp)) {
    if (strncmp (line, DIFF3_MERGE_PAT_3, sizeof DIFF3_MERGE_PAT_3 - 1) == 0) {
      result = 1;
      goto out;
    }
    while (line [strlen (line)-1] != '\n') {
      // warn << "long line\n";
      if (!fgets (line, 2048, fp))
	break;
    }
  }
  if (ferror (fp))
    return -1;

out:
  fclose (fp);
  return result;
}

