/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include "pastwatch.h"
#include "opts.h"
#include "path.h"
#include "pdb.h"
#include "io.h"

extern ptr<cfsrw> _cfsrw;

extern int  check_dirent (str dirpath, dirent *e, bool& ignore_entry);
extern void read_file (int, bigint, bigint, callback<void, cfsrw_status>::ref);
extern int  diff_context (const char *f1, const char *f2, strbuf &b, str t);
extern int  diff   (const char *f1, const char *f2, strbuf &b);
extern int  isdiff (const char *f1, const char *f2);
extern int  isconf (const char *path);
extern str  get_tag_of_commit (bigint cn);

#define ISDIR(sb) ((sb.st_mode & S_IFDIR) == S_IFDIR)

class diff_object : public virtual refcount {
private:
  vec<str> objects;
  str path;
  cbi cb;
  unsigned rvn;
  bigint rv;
  bigint fh;
  bigint rep_saddr;
  bigint tag_saddr;

  void
  read_done (str tfn, int tfd, cfsrw_status status)
  {
    close (tfd);
    if (status) {
      unlink (tfn);
      warn << "cannot read version " << arg_revision0 << " of "
	   << path << " from repository\n";
      next ();
      return;
    }

    if (isdiff (tfn.cstr (), path.cstr ()) > 0) {
      str rt = get_tag_of_commit (rv);
      strbuf out;
      out << "File: " << path << "\n";
      out << "===============================================" << "\n";
      out << "Repository version: " << rt << "\n";
      if (arg_contextdiff)
        out << "diff -c -t " << rt << " " << path << "\n";
      else
        out << "diff -t " << rt << " " << path << "\n";
      out.tosuio ()->output (1);

      str opts = "";
      strbuf b;
      if (arg_contextdiff) {
	strbuf x;
	str n;
        get_parent_dir (path, n);
	x << "r" << rvn << " of " << n;
	diff_context (tfn.cstr (), path.cstr (), b, x);
      }
      else
	diff (tfn.cstr (), path.cstr (), b);
      b.tosuio ()->output (1);
    }

    unlink (tfn);
    next ();
  }





  void
  getattr (cfsrcallp c, cfsrw_status status)
  {
    if (status) {
      warn << "cannot get attribute of " << path << "\n";
      next ();
      return;
    }
    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (c->res ());
    rvn = res->attr->revision;
    rv = res->attr->commit;

    str s = pastwatch_tmpfile ();
    int fd = open (s, O_CREAT | O_TRUNC | O_WRONLY, 0600);
    read_file
      (fd, fh, rep_saddr, wrap (mkref (this), &diff_object::read_done, s, fd));
  }

  void
  find_fh ()
  {
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR, wrap (mkref (this), &diff_object::getattr));
    cfsr_getattr_arg *a = static_cast<cfsr_getattr_arg *> (c->arg ());
    a->object = fh;
    cfsr_call (rep_saddr, c);
  }

  void
  findver_cb (cfsrcallp c, cfsrw_status status)
  {
    if (status) {
      warn << "cannot find revision " << arg_revision0
	   << " of " << path << "\n";
      next ();
      return;
    }
    cfsr_findver_res *res = static_cast<cfsr_findver_res *> (c->res ());
    rep_saddr = res->saddr;
    find_fh ();
  }

  void
  expand (str path)
  {
    DIR *dir = opendir (path);
    if (dir == 0)
      return;

    struct dirent *e;
    while ((e = readdir (dir))) {
      bool should_ignore = false;
      if (check_dirent (path, e, should_ignore))
	continue;
      if (!should_ignore) {
	str fp;
	if (path == ".")
	  fp = e->d_name;
	else
	  fp = path << "/" << e->d_name;
	if (e->d_type == DT_DIR && arg_recurse)
	  expand (fp);
	else
	  objects.push_back (fp);
      }
    }
    closedir (dir);
  }

  void
  next () 
  {
    if (objects.size () == 0) {
      cb (0);
      return;
    }

    path = objects.pop_front ();
    ptr<pfe> v = pmgr->lookup (path);
    if (!v) {
      warn << "no information for " << path << ", not checked-in?\n";
      next ();
      return;
    }
    if (isconf (path) != 0) {
      warn << path << " contains conflicts\n";
      next ();
      return;
    }

    fh = v->fh;
    if (arg_tag0 != "")
      rep_saddr = tag_saddr;
    else
      rep_saddr = my_branch_head_snapshot (_cfsrw->getreplica ());

    if (arg_revision0) {
      cfsrcallp c = New refcounted<cfsrcall>
	(CFSRPROC_FINDVER, wrap (mkref (this), &diff_object::findver_cb));
      cfsr_findver_arg *a = static_cast<cfsr_findver_arg *> (c->arg ());
      a->object = fh;
      a->revision = arg_revision0;
      cfsr_call (rep_saddr, c);
      return;
    }
    else {
      if (arg_tag0 == "" && v->sticky != 0)
        rep_saddr = v->sticky;
      find_fh ();
    }
  }

public:
  diff_object (str path, cbi cb)
    : path (path), cb (cb)
  {
    //if (arg_tag0)

    struct stat sb;
    if (lstat (path, &sb)) {
      warn << "cannot open " << path << "\n";
      cb (-1);
      return;
    }
    else if (ISDIR (sb))
      expand (path);
    else
      objects.push_back (path);

    if (arg_tag0 != "") {
      bigint n;
      dearmor64A_bigint (arg_tag0, n);
      ptr<btree_srv> srv = _cfsrw->getreplica ()->btree ();
      btree_node x;
      srv->lookup_by_commit (n, x);
      tag_saddr = x.commit.vptr;
    }

    next ();
  }
};

void
diff (bigint root, str path, cbi cb)
{
  ptr<diff_object> a = New refcounted<diff_object> (path, cb);
}

