/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <dirent.h>
#include "rxx.h"
#include "qhash.h"
#include "btree.h"
#include "stdinc.h"
#include "opts.h"
#include "path.h"

extern "C" {
time_t get_date (char *raw);
};

const char* cmd_list[] = {
  "import", "commit", "rename", "rtag", "branch",
  "history", "tags", "latest", "fetch", "checkout",
  "merge", "update", "up", "add", "remove", "clear",
  "status", "diff", "log", "rescue"
};
const unsigned num_commnds = 20;

const int cmd_import = 0;
const int cmd_checkout = 1;
const int cmd_update = 2;
const int cmd_commit = 3;
const int cmd_add = 4;
const int cmd_remove = 5;
const int cmd_rename = 6;
const int cmd_clear = 7;
const int cmd_diff = 8;
const int cmd_status = 9;
const int cmd_log = 10;
const int cmd_rescue = 11;
const int cmd_branch = 12;
const int cmd_rtag = 13;
const int cmd_fetch = 14;
const int cmd_merge = 15;
const int cmd_latest = 16;
const int cmd_history = 17;
const int cmd_tags = 18;

str arg_localpath = "";
str arg_branchtag = "";
str arg_jointag = "";
str arg_branchfrom = "";
str arg_viewstr;
str arg_keyfile;
str arg_keystr;
str arg_branch;
str arg_message = "";
str arg_tag0 = "";
str arg_tag1 = "";
str arg_date0 = "";
str arg_date1 = "";
str arg_timestamp = "";
str arg_newtag = "";

int arg_command = -1;
int arg_debugmode = 0;
int arg_revision0 = 0;
int arg_revision1 = 0;

bool arg_noupdate = false;
bool arg_resetsticky = false;
bool arg_readonly = true;
bool arg_updatedir = false;
bool arg_checkonly = false;
bool arg_force = false;
bool arg_contextdiff = false;
bool arg_relink = false;
bool arg_dironly = false;
bool arg_recurse = true;
bool arg_killbranch = false;
bool arg_newbranch = false;
bool arg_notrans = false;
bool arg_ignore_implicit_branches = false;

bigint arg_view;
bigint arg_key;
bigint arg_handle;

vec<str> arg_views;
vec<str> arg_keys;
vec<str> arg_branches;
vec<ptr<vec<str> > > arg_paths;

extern int  check_dirent (str dirpath, dirent *e, bool& ignore_entry);
extern void check_skip   (str dirpath, str name, bool& ignore_entry);
extern void list_branches ();

#define ATOI(s,v) \
{ \
  char *end = 0; \
  v = strtol (s, &end, 10); \
  if (*end != '\0') { \
    usage (); \
    return 1; \
  } \
}

bool
is_cmd(char *s) {
  for(unsigned i=0; i<num_commnds; i++)
    if (!strncmp(s, cmd_list[i], strlen(cmd_list[i]))) 
      return true;
  return false;
}

str
get_btag_from_infofile (str path)
{
  str x;
  str p = get_info_file (get_parent_dir (path));
  str buf = file2str (p);
  if (!buf) {
    p = get_info_file (get_dir (path));
    buf = file2str (p);
  }
  if (buf) {
    vec<str> n;
    split (&n, "\n", buf);
    assert (n.size () >= 4);
    if (n.size () == 4)
      return x;
    else
      return n [4];
  }
  return x;
}

str
get_commit_record_from_infofile (str path)
{
  str p = get_info_file (get_parent_dir (path));
  str buf = file2str (p);
  if (!buf) {
    p = get_info_file (get_dir (path));
    buf = file2str (p);
  }
  if (buf) {
    vec<str> n;
    split (&n, "\n", buf);

    if (n.size () < 4) {
      warn << "Error in Pastwatch/info file.  File does not contain branch field.\n";
      assert(0);
    } 
    return n [3];
  }
  return NULL;
}

static int
get_info (str path, char *sk_file, char *viewstr, char *branch)
{
  str kstr, vstr, bstr;

  if (sk_file)
    kstr = str (sk_file, strlen (sk_file));
  if (viewstr)
    vstr = str (viewstr, strlen (viewstr));
  if (branch)
    bstr = str (branch, strlen (branch));

  if (!kstr || !vstr || !bstr) {
    str p = get_info_file (get_parent_dir (path));
    str buf = file2str (p);
    if (!buf) {
      p = get_info_file (get_dir (path));
      buf = file2str (p);
    }
    if (buf) {
      vec<str> n;
      split (&n, "\n", buf);
      assert (n.size () >= 4);

      if (!kstr)
	kstr = n [1];
	
      if (!vstr)
	vstr = n [2];
    
      if (!bstr) {
	bstr = n [4];
	if (bstr && bstr != "") 
	  warn << "Using branch: " << bstr << "\n";
      }
    }
  }

  if (!kstr || !vstr)
    return -1;

  if (!bstr)
    bstr = "";
 
  char sym[1024];
  rxx rx ("project:([0-9a-zA-Z+-]*)[/]*$");
  int r = readlink (vstr.cstr (), sym, 1024);
  if (r > 0) sym[r] = '\0';
  str v;
  if (rx.search (vstr))
    v = rx [1];
  else if (r > 0 && rx.search (sym))
    v = rx [1];
  else
    v = vstr;

  for (unsigned i=0; i<arg_views.size (); i++) {
    if (arg_views [i] == vstr &&
	arg_keys [i] == kstr &&
	arg_branches [i] == bstr) {
      ptr<vec<str> > z = arg_paths [i];
      z->push_back (path);
      arg_paths [i] = z;
      return 0;
    }
  }

  arg_views.push_back (v);
  arg_keys.push_back (kstr);
  arg_branches.push_back (bstr);
  ptr<vec<str> > z = New refcounted<vec<str> >;
  z->push_back (path);
  arg_paths.push_back (z);
  return 0;
}

static void
usage ()
{
  warnx << "Pastwatch version " << VERSION << "\n";
  warnx << "usage: " << progname << " [opts] cmd [cmd opts] files\n";
  warnx << " -n: check only, don't make changes\n";
  warnx << " -l: don't open sub-directories\n";
  warnx << " -b <branch tag>: specify a branch to use\n";
  warnx << " -x <n>: debug mask\n";
  warnx << "   1 - reports latency to fetch project block and public keys\n";
  warnx << "   2 - reports latency to fetch blocks\n";
  warnx << "   4 - outputs debug messages when updating a directory\n";
  warnx << " -s <file>: public/private key file\n";
  warnx << " -p <project id>: project identification number\n";
  warnx << " See the man page for commands and command options: pastwatch(1).\n";
}

int update_debug = 0;
int ns_debug = 0;
int cfs_debug = 0;
int view_debug = 0;

int
getopts (int argc, char **argv)
{
  char *sk_file = 0;
  char *viewstr = 0;
  char *branch = 0;
  int    pastargc = 0;

  if (argc == 1) {
    usage ();
    return 1;
  }

  for(int i=0; i<argc; i++) {
    if (is_cmd(argv[i])) {
      pastargc = i;
      break;
    }
  }

  int ch;
  while ((ch = getopt (pastargc, argv, "nlb:x:s:p:i")) != -1) {
    switch (ch) {
    case 'b':
      branch = optarg;
      break;
    case 'x':
      ATOI (optarg, arg_debugmode);
      break;
    case 'l':
      arg_recurse = false;
      break;
    case 'n':
      arg_checkonly = true;
      break;
    case 'p':
      viewstr = optarg;
      break;
    case 's':
      sk_file = optarg;
      break;
    case 'i':
      arg_ignore_implicit_branches = true;
      break;
    case '?':
    default:
      usage ();
      return 1;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc == 0) {
    usage ();
    return 1;
  }

  if (arg_debugmode & 1)
    view_debug = 1;
  if (arg_debugmode & 2)
    ns_debug = 1;
  if (arg_debugmode & 4)
    update_debug = 1;

  if (strncmp (argv[0], "import", 6) == 0) {
    arg_readonly = false;
    arg_command = cmd_import;
  }
  else if (strncmp (argv[0], "commit", 6) == 0) {
    arg_readonly = false;
    arg_command = cmd_commit;
  }
  else if (strncmp (argv[0], "rename", 6) == 0) {
    arg_readonly = false;
    arg_command = cmd_rename;
  }
  else if (strncmp (argv[0], "rtag", 4) == 0) {
    arg_checkonly = true;
    arg_readonly = false;
    arg_command = cmd_rtag;
  }
  else if (strncmp (argv[0], "branch", 6) == 0) {
    arg_checkonly = true;
    arg_readonly = false;
    arg_command = cmd_branch;
  }
  else if (strncmp (argv[0], "history", 7) == 0) {
    arg_checkonly = true;
    arg_command = cmd_history;
  }
  else if (strncmp (argv[0], "tags", 4) == 0) {
    arg_checkonly = true;
    arg_command = cmd_tags;
  }
  else if (strncmp (argv[0], "latest", 6) == 0) {
    arg_checkonly = true;
    arg_command = cmd_latest;
  }
  else if (strncmp (argv[0], "fetch", 5) == 0) {
    arg_checkonly = true;
    arg_command = cmd_fetch;
  }

  else if (strncmp (argv[0], "checkout", 8) == 0)
    arg_command = cmd_checkout;
  
  else if (strncmp (argv[0], "merge", 5) == 0)
    arg_command = cmd_merge;

  else if (strncmp (argv[0], "update", 6) == 0)
    arg_command = cmd_update;

  else if (strncmp (argv[0], "up", 2) == 0)
    arg_command = cmd_update;

  else if (strncmp (argv[0], "add", 3) == 0)
    arg_command = cmd_add;

  else if (strncmp (argv[0], "remove", 6) == 0)
    arg_command = cmd_remove;

  else if (strncmp (argv[0], "clear", 5) == 0)
    arg_command = cmd_clear;

  else if (strncmp (argv[0], "status", 6) == 0)
    arg_command = cmd_status;

  else if (strncmp (argv[0], "diff", 4) == 0)
    arg_command = cmd_diff;

  else if (strncmp (argv[0], "log", 3) == 0)
    arg_command = cmd_log;

  else if (strncmp (argv[0], "rescue", 6) == 0)
    arg_command = cmd_rescue;

  // optreset = 1;
  optind = 1;

  //
  // add -S timestamp
  //
  if (arg_command == cmd_add) {
    while ((ch = getopt (argc, argv, "S:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // remove -f -d
  //
  else if (arg_command == cmd_remove) {
    while ((ch = getopt (argc, argv, "df")) != -1)
      switch (ch) {
      case 'd':
	arg_dironly = true;
	break;
      case 'f':
        arg_force = true;
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // merge -d -t Tag
  //
  else if (arg_command == cmd_merge) {
    arg_noupdate = true;
    while ((ch = getopt (argc, argv, "dt:")) != -1)
      switch (ch) {
      case 'd':
        arg_updatedir = true;
        break;
      case 't':
	arg_tag0 = str (optarg, strlen (optarg));
	break;
      default:
	usage ();
	return 1;
      }
    if (arg_tag0 == "") {
      warn << "merge command requires a version\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // update -A -d -D Date -r Rev -t Tag
  //
  else if (arg_command == cmd_update) {
    while ((ch = getopt (argc, argv, "AdD:r:t:")) != -1)
      switch (ch) {
      case 'A':
	arg_resetsticky = true;
	break;
      case 'd':
        arg_updatedir = true;
        break;
      case 'D':
	arg_date0 = str (optarg, strlen (optarg));
	break;
      case 't':
	arg_tag0 = str (optarg, strlen (optarg));
	break;
      case 'r':
	ATOI (optarg, arg_revision0);
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // checkout -t Tag -d Dir
  //
  else if (arg_command == cmd_checkout) {
    while ((ch = getopt (argc, argv, "t:d:")) != -1)
      switch (ch) {
      case 't':
	arg_tag0 = str (optarg, strlen (optarg));
	break;
      case 'd':
        arg_localpath = str (optarg, strlen (optarg));
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // rtag -S timestamp -D date -t Tag Tag
  //
  else if (arg_command == cmd_rtag) {
    while ((ch = getopt (argc, argv, "S:D:t:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'D':
	arg_date0 = str (optarg, strlen (optarg));
	break;
      case 't':
	arg_tag0 = str (optarg, strlen (optarg));
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // diff -c -D Date -t Tag -r Rev
  //
  else if (arg_command == cmd_diff) {
    while ((ch = getopt (argc, argv, "D:cr:t:")) != -1)
      switch (ch) {
      case 'D':
	arg_date0 = str (optarg, strlen (optarg));
	break;
      case 'c':
        arg_contextdiff = true;
        break;
      case 't':
	arg_tag0 = str (optarg, strlen (optarg));
	break;
      case 'r':
	ATOI (optarg, arg_revision0);
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // branch -S timestamp -k -B -f tag -b str -m Msg
  //
  else if (arg_command == cmd_branch) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:kBf:b:m:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'B':
	arg_newbranch = true;
	break;
      case 'k':
	arg_killbranch = true;
	break;
      case 'f':
	arg_newbranch = true;
	arg_branchfrom = str (optarg, strlen (optarg));
	break;
      case 'b':
	arg_branchtag = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    if (arg_newbranch && arg_killbranch) {
      warn << "-k and -B are incompatible options\n";
      usage ();
      return 1;
    }
    if (arg_branchtag && arg_branchtag != "" && arg_killbranch) {
      warn << "-k and -b are incompatible options\n";
      usage ();
      return 1;
    }
    if (arg_branchfrom && arg_branchfrom != "" && arg_killbranch) {
      warn << "-f and -k are incompatible options\n";
      usage ();
      return 1;
    }
    if (!has_msg) {
      warn << "must provide a commit message using -m\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // commit -S timestamp -j tag -m Msg
  //
  else if (arg_command == cmd_commit) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:m:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'j':
	arg_jointag = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    if (!has_msg) {
      warn << "must provide a commit message using -m\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // rename -S timestamp -m Msg
  //
  else if (arg_command == cmd_rename) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:m:")) != -1)
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // import -S timestamp -m Msg
  //
  else if (arg_command == cmd_import) {
    bool has_msg = false;
    while ((ch = getopt (argc, argv, "S:m:")) != -1) {
      switch (ch) {
      case 'S':
	arg_timestamp = str (optarg, strlen (optarg));
	break;
      case 'm':
	arg_message = str (optarg, strlen (optarg));
	has_msg = true;
        break;
      default:
	usage ();
	return 1;
      }
    }
    if (!has_msg) {
      warn << "must provide a commit message using -m\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // rescue -a -h Handle
  //
  else if (arg_command == cmd_rescue) {
    bool has_handle = false;
    while ((ch = getopt (argc, argv, "ah:")) != -1)
      switch (ch) {
      case 'a':
	arg_relink = true;
	break;
      case 'h':
	{
          str x = dearmor64A (str (optarg, strlen (optarg)));
          mpz_set_rawmag_be (&arg_handle, x.cstr(), x.len());
	  has_handle = true;
	}
        break;
      default:
	usage ();
	return 1;
      }
    if (!has_handle) {
      warn << "must specify a handle to rescue using -h\n";
      usage ();
      return 1;
    }
    argc -= optind;
    argv += optind;
  }
  //
  // log -d
  //
  else if (arg_command == cmd_log) {
    while ((ch = getopt (argc, argv, "d")) != -1)
      switch (ch) {
      case 'd':
	arg_dironly = true;
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }
  //
  // status -d
  //
  else if (arg_command == cmd_status) {
    while ((ch = getopt (argc, argv, "d")) != -1)
      switch (ch) {
      case 'd':
	arg_dironly = true;
	break;
      default:
	usage ();
	return 1;
      }
    argc -= optind;
    argv += optind;
  }

  else {
    argc -= 1;
    argv += 1;
  }

  if (arg_command == -1) {
    usage ();
    return 1;
  }

  if (arg_revision0 != 0 && arg_tag0 && arg_tag0 != "") {
    warn << "cannot accept both a revision and a tag\n";
    usage ();
    return 1;
  }

  if (arg_date0 && arg_date0 != "" && arg_revision0 != 0) {
    warn << "cannot accept both a revision and a date\n";
    usage ();
    return 1;
  }
  
  if (arg_date0 && arg_date0 != "" && arg_tag0 && arg_tag0 != "") {
    warn << "cannot accept both a tag and a date\n";
    usage ();
    return 1;
  }

  if (arg_date0 && arg_date0 != "") {
    time_t t = get_date ((char *)arg_date0.cstr ());
    if (t < 0) {
      warn << "cannot parse " << arg_date0 << "\n";
      usage ();
      return 1;
    }
  }

  vec<str> paths;

  if (argc == 0) {
    if (arg_command == cmd_import || arg_command == cmd_checkout ||
	arg_command == cmd_add || arg_command == cmd_rescue) {
      warn << "must provide a file or directory name\n";
      usage ();
      return 1;
    }
    else if (arg_command == cmd_rtag) {
      warn << "must provide a tag\n";
      usage ();
      return 1;
    }
    paths.push_back (".");
  }
  else if (argc > 1) {
    if (arg_command == cmd_import || arg_command == cmd_checkout ||
	arg_command == cmd_rescue) {
      warn << "please give only one file name\n";
      usage ();
      return 1;
    }
    else if (arg_command == cmd_rtag) {
      warn << "rtag command only takes one argument\n";
      usage ();
      return 1;
    }
  }
  else if (argc == 1) {
    str x = sanitize_path (argv [0]);
    if (arg_command == cmd_import && (x == "." || x == "..")) {
      warn << "import - invalid module name\n";
      return 1;
    }
    if (arg_command == cmd_checkout && (x == "/")) {
      warn << "checkout - invalid module name\n";
      return 1;
    }
  }

  if (arg_command == cmd_import || arg_command == cmd_checkout) {
    if (!branch)
      branch = "";
    if (!sk_file)
      sk_file = "";
    if (!viewstr) {
      warn << "no project id\n";
      usage ();
      return 1;
    }
  }

  for (int i=0; i<argc; i++) {
    str x = sanitize_path (argv [i]);
    paths.push_back (x);
  }

  while (!paths.empty ()) {
    str path = paths.pop_front ();
    struct stat sb;
    int err = lstat (path, &sb);
  
    int r = get_info (path, sk_file, viewstr, branch);
    if (r > 0)
      return r;

    // these commands do not need files, so don't recurse in the
    // current directory. if can't find a view or key, just stop.
    else if (r < 0 && (arg_command == cmd_branch ||
	               arg_command == cmd_rtag ||
	               arg_command == cmd_fetch ||
	               arg_command == cmd_latest ||
	               arg_command == cmd_tags ||
		       arg_command == cmd_history)) {
      warn << "no key or project id\n";
      return 1;
    }

    // r < 0 iff the current directory does not contain a Pastwatch
    // directory. in this case we recurse.
    else if (r < 0 && !err && (sb.st_mode & S_IFDIR)) {
      DIR *dir = opendir (path);
      if (!dir)
        warn << "cannot open " << path << "\n";
      else {
        struct dirent *e; 
        while ((e = readdir (dir))) {
          bool blah = false;
          if (check_dirent (path, e, blah))
	    continue;
	  bool skip = false;
	  check_skip (path, str (e->d_name), skip);
	  if (skip)
	    continue;
          str p = path << "/" << e->d_name;
          get_info (p, sk_file, viewstr, branch);
        }
        closedir (dir);
      }
    }
    else if (r != 0) {
      warn << "no key or project id for " << path << "\n";
      return 1;
    }
  }

  if (arg_command == cmd_rename) {
    if (arg_views.size () != 1) {
      warn << "rename command needs two files from the same project\n";
      usage ();
      return 1;
    }
    if (arg_paths [0]->size () != 2) {
      warn << "rename command needs two files from the same project\n";
      usage ();
      return 1;
    }
  }

  return 0;
}

str
cmd2str (int cmd)
{
  str s;
  if (cmd == cmd_checkout)
    s = "checkout";
  else if (cmd == cmd_merge)
    s = "merge";
  else if (cmd == cmd_update)
    s = "update";
  else if (cmd == cmd_import)
    s = "import";
  else if (cmd == cmd_commit)
    s = "commit";
  else if (cmd == cmd_add)
    s = "add";
  else if (cmd == cmd_remove)
    s = "remove";
  else if (cmd == cmd_clear)
    s = "clear";
  else if (cmd == cmd_log)
    s = "log";
  else if (cmd == cmd_status)
    s = "status";
  else if (cmd == cmd_diff)
    s = "diff";
  else if (cmd == cmd_rescue)
    s = "rescue";
  else if (cmd == cmd_rename)
    s = "rename";
  else if (cmd == cmd_tags)
    s = "tags";
  else if (cmd == cmd_history)
    s = "history";
  else if (cmd == cmd_latest)
    s = "latest";
  else if (cmd == cmd_fetch)
    s = "fetch";
  else if (cmd == cmd_branch)
    s = "branch";
  else if (cmd == cmd_rtag)
    s = "rtag";
  else
    assert (0);
  return s;
}

