/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef PASTWATCH_PDB_H
#define PASTWATCH_PDB_H

#include <qhash.h>
#include "cfsw_prot.h"
#include "adb.h"

// entries in the Pastwatch db file

class pfe : public adb_entry 
{
public:
  str name;
  bigint fh;
  bigint saddr;
  mtime_t mtime;
  bigint sticky;

  pfe () {}
  virtual ~pfe () {}

  pfe (bigint h) 
    : fh (h)
  {
    saddr = 0;
    sticky = 0;
    mtime.sec = 0; mtime.nsec = 0;
  }

  pfe (bigint h, mtime_t m)
    : fh (h), saddr (0), mtime (m), sticky (0)
  {
  }

  pfe (bigint h, bigint s)
    : fh (h), saddr (s), sticky (0)
  {
    mtime.sec = 0; mtime.nsec = 0;
  }

  pfe (bigint h, bigint s, bigint t)
    : fh (h), saddr (s), sticky (t)
  {
    mtime.sec = 0; mtime.nsec = 0;
  }

  pfe (bigint h, bigint s, mtime_t m)
    : fh (h), saddr (s), mtime (m), sticky (0)
  {
  }

  pfe (bigint h, bigint s, mtime_t m, bigint t)
    : fh (h), saddr (s), mtime (m), sticky (t)
  {
  }

  bool match (str n)
  {
    if (name == n)
      return true;
    return false;
  }

  void print_entry (strbuf &b);
  bool parse_entry (const char *);
};

class pdb_mgr {
private:
  int insert (str path, pfe);
  ptr<adb> opendb (str dir);

  vec<str> add_path;
  vec<pfe> add_pfe;
  vec<str> mkdir_path;
  vec<str> mkdir_cfsp;
  vec<bool> mkdir_renamed;
  vec<str> remove_path;
  vec<str> opened_dirs;
  vec<str> supers_dirs;
  vec<bigint> supers_commits;

  vec<ptr<adb> > adbs;
  qhash<str, ptr<adb> > adbhash;

public:
  pdb_mgr () {}
  ~pdb_mgr () {}
  int init ();

  int process (bigint s, str btag);
  bool exists (str path);
  ptr<pfe> lookup (str path);
  ptr<pfe> lookup_in_parent_dir (str path);

  void schedule_mkdir (str path, str cfspath);
  void schedule_add (str path, pfe);
  void schedule_supersedes (str path, bigint n);
  void schedule_remove (str path);
  void schedule_rename (str newdir, str cfsto);

  int remove (str path);
  int remove_dir (str path);
};

extern pdb_mgr* pmgr;


class dirop : public adb_entry {
public:
  char which;
  str name;
  bigint fh;

  dirop (char w) : which (w), name (""), fh (0) {}
  dirop (char w, str name) : which (w), name (name), fh (0) {}
  dirop (char w, str name, bigint fh) : which (w), name (name), fh (fh) {}
  virtual ~dirop () {}

  str getkey ()
  {
    str x;
    if (which == 'A')
      x = "A " << name;
    else if (which == 'R')
      x = "R " << name;
    else
      assert (0);
    return x;
  }
  
  bool match (str n)
  {
    str x = getkey ();
    if (x == n)
      return true;
    return false;
  }

  virtual void print_entry (strbuf &b) = 0;
  virtual bool parse_entry (const char *) = 0;
};

class add_dirop : public dirop {
public:
  add_dirop () : dirop ('A') {}
  add_dirop (str name) : dirop ('A', name) { }
  add_dirop (str name, bigint fh) : dirop ('A', name, fh) { }

  void print_entry (strbuf &b) {
    if (fh != 0)
      b << "A " << name << "/" << armor64A (fh);
    else
      b << "A " << name;
  }
  bool parse_entry (const char *);
};

class remove_dirop : public dirop {
public:
  bigint saddr;
  remove_dirop () : dirop ('R') {}
  remove_dirop (str name, bigint fh, bigint s)
    : dirop ('R', name, fh), saddr (s) { }

  void print_entry (strbuf &b) {
    b << "R " << name << "/" << armor64A (fh)
      << "/" << armor64A (saddr);
  }
  bool parse_entry (const char *);
};

class dirop_factory : public adb_entry_factory {
public:
  ptr<adb_entry>
  parse_entry (const char *s) {
    if (s[0] == 'A') {
      ptr<add_dirop> a = New refcounted<add_dirop> ();
      a->which = 'A';
      if (!a->parse_entry (s)) {
        warn << "skipping malformed line " << s << "\n";
	return 0;
      }	
      else
        return a;
    }
    else if (s[0] == 'R') {
      ptr<remove_dirop> a = New refcounted<remove_dirop> ();
      a->which = 'R';
      if (!a->parse_entry (s)) {
        warn << "skipping malformed line " << s << "\n";
	return 0;
      }
      else
        return a;
    }
    else {
      warn << "skipping malformed line " << s << "\n";
      return 0;
    }
  }
};

struct dirops
{
  vec<ptr<add_dirop> > add;
  vec<ptr<remove_dirop> > remove;

  dirops () {}

  void clear () {
    add.clear ();
    remove.clear ();
  }

  void push_add (ptr<add_dirop> a) { add.push_back (a); }
  void push_add (dirops &d, unsigned i) { push_add (d.add [i]); }
  void copy_add (dirops &d) { add = d.add; }

  void push_remove (ptr<remove_dirop> r) { remove.push_back (r); }
  void push_remove (dirops &d, unsigned i) { push_remove (d.remove [i]); }
  void copy_remove (dirops &d) { remove = d.remove; }
};

extern int new_dirop  (str dir, ptr<dirop>);
extern int del_dirop  (str path, char w);
extern int set_dirops (str dir, struct dirops &d);
extern int get_dirops (str dir, struct dirops &d);
extern int del_dirops (const vec<str> &, const vec<char> &);

#endif

