
/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/time.h>
#include "stdinc.h"
#include "parseopt.h"
#include "qhash.h"
#include "rxx.h"
#include <dhash_common.h>
#include "ns.h"
#include "vervec.h"
#include "traverse.h"
#include "view.h"
#include "cfsw_prot.h"
#include "record.h"
#include "cfsw_record.h"
#include "snapshot.h"
#include "cfs.h"
#include "pastupd.h"

extern void sync_resched ();
extern void sync_repository (vec<str> n);

extern void iterate_write
(deltap r, callback<void,bigint,cfsrw_status>::ref cb);

int view_debug = 0;
int ns_debug = 0;
int cfs_debug = 0;

class syncer {
private:
  ptr<view> _view;
  bigint _vid;
  int _work;
  int _err;
  vec<bigint> _newpks;
  vec<bigint> _inlogs;
  vec<bigint> _depends;
  vec<bigint> _stop;
  comvec _commits;
  ptr<ns_block> _insert_this_commit;
  bigint _insert_this_commit_hash;
  qhash<bigint, bool> _done;
  bool _ispush;

  vec<deltap> _deltas;
  vec<ptr<ns_block> > _delta_blocks;
  vec<bigint> _delta_hashes;
  unsigned _ndeltas;
  
  vec<str> _rlist;

  void
  insert_pks_cb (ns_stat status, bigint n)
  {
    if (status) {
      warn << "cannot insert pk " << armor64A (n) << " into DHT\n";
      nsobj = 0;
      sync_repository (_rlist);
      delete this;
    }
    else
      write_pks ();
  }

  void
  write_pks ()
  {
    if (_newpks.size ()) {
      bigint n = _newpks.pop_front ();
      warn << "moving pk " << armor64A (n) << "\n";
      nsobj->insert_raw_pk_from_cache
	(n, wrap (this, &syncer::insert_pks_cb));
    }
    else {
      nsobj = 0;
      sync_repository (_rlist);
      delete this;
    }
  }

  void
  finalize ()
  {
    if (_work == 0) {
      if (_err == 0)
	write_pks ();
      else {
        nsobj = 0;
        sync_repository (_rlist);
        delete this;
      }
    }
  }

  void
  insert_done (ns_stat status, bigint n)
  {
    _work--;
    if (status) {
      warn << "cannot insert " << armor64A (n) << "\n";
      _err++;
      finalize ();
      return;
    }
    if (_work == 0)
      check_commits ();
  }

  void
  get_write_block (bigint n, ptr<ns_block> b)
  {
    _work--;
    if (!b) {
      warn << "cannot fetch write block!\n";
      _err++;
      finalize ();
    }
    else {
      _work++;
      bigint x = nsobj->insert_block
	(b->data, b->len, wrap (this, &syncer::insert_done), true);
      warn << "block inserted: " << x << "\n";
      assert (x == n);
    }
  }

  void
  dump_write_cb (bigint n, cfsrw_status status)
  {
    if (!status) {
      _work++;
      get_write_block (n, nsobj->retrieve_local (n));
    }
    else {
      if (status != CFSRWERR_NOMORELOG) {
        warn << "iterate write failed!\n";
	_err++;
      }
      _work--;
      finalize ();
    }
  }
  
  void
  dump_write (deltap r)
  {
    _work++;
    iterate_write (r, wrap (this, &syncer::dump_write_cb));
  }
  
  void
  fetch_deltap_cb (bigint x, unsigned i, ptr<ns_block> b)
  {
    _work--;
    
    _ndeltas++;
    if (_ndeltas == _deltas.size ())
      _work--;

    if (!b) {
      warn << "cannot load record " << armor64A (x) << "\n";
      _err = 1;
      finalize ();
      return;
    }
  
    str s (b->data, b->len);
    deltap r = cfsw_deltap_parse (s);
    if (!r) {
      warn << "cannot load record " << armor64A (x) << "\n";
      _err = 1;
      finalize ();
      return;
    }

    _deltas [i] = r;
    _delta_blocks [i] = b;
  
    if (_ndeltas == _deltas.size ()) {
      for (unsigned i=0; i<_deltas.size (); i++) {
        _work++;
        bigint xx = nsobj->insert_block
          (_delta_blocks [i]->data, _delta_blocks [i]->len,
	   wrap (this, &syncer::insert_done), true);
	warn << "delta inserted: " << xx << "\n";
	assert (xx == _delta_hashes [i]);
        if (_deltas [i]->rtype () == cfsw_delta::T_WRITE)
          dump_write (_deltas [i]);
      }
    }
  }

  void
  fetch_ind_cb (bigint n, ptr<ns_block> b)
  {
    _work--;
    if (!b) {
      warn << "cannot load indirect " << armor64A (n) << "\n";
      _err = 1;
      finalize ();
      return;
    }
    _work++;
    nsobj->insert_block
      (b->data, b->len, wrap (this, &syncer::insert_done), true);
    warn << "indirect inserted: " << n << "\n";
  }

  void
  get_ind_cb (crec, bigint n)
  {
    _work++;
    nsobj->retrieve_block (n, wrap (this, &syncer::fetch_ind_cb, n));
  }
  
  void
  get_deltas_cb (crec c, bigint n, bool failed)
  {
    if (failed) {
      _work--;
      warn << "cannot iterate log in " << armor64A (n) << "\n";
      _err = 1;
      finalize ();
      return;
    }

    unsigned i = _delta_hashes.size ();
    _delta_hashes.push_back (n);
    _work++;
    nsobj->retrieve_block (n, wrap (this, &syncer::fetch_deltap_cb, n, i));
  }

  void
  get_cached_commit (bool inlog, bigint c, bool dht, ptr<ns_block> b)
  {
    if (!b) { // yikes
      warn << "cannot fetch " << armor64A (c) << " from replica!\n";
      _err = 1;
      finalize ();
      return;
    }

    crec r = commit_record::alloc (str (b->data, b->len));
    if (!r) { // yikes
      warn << "cannot parse " << armor64A (c) << " into commit record!\n";
      _err = 1;
      finalize ();
      return;
    }

    bool haslog = false;
    unsigned logidx = 0;
    for (unsigned i=0; i<_commits.size () && !haslog; i++) {
      if (r->depends ().head == _commits.head (i)) {
	logidx = i;
	commit_info ci = r->depends();
	warn << armor64A (ci.rptr) << " is in a known log\n";
	haslog = true;
      }
    }

    if (inlog && r->prev () != 0) {
      assert (haslog);
      if (r->prev () != _stop [logidx]) {
        bool *x = _done [r->prev ()];
        if (!x) {
	  _inlogs.push_back (r->prev ());
	  _done.insert (r->prev (), true);
        }
      }
    }

    if (r->depends ().rptr != 0) {
      if (!haslog) {
        bool *x = _done [r->depends ().rptr];
        if (!x) {
	  commit_info ci = r->depends();
	  warn << armor64A (ci.rptr) << " in unknown log\n";

	  if (!inlog && dht && !_ispush) {
	    // if this block itself is not in a known log, and it is
	    // in the DHT, and we are not in push mode, then we assume
	    // blocks it depends on are in the DHT as well
	  }
	  else {
	    _depends.push_back (r->depends ().rptr);
	    _done.insert (r->depends ().rptr, true);
	  }
        }
      }
    }

    if (!dht || _ispush) {
      warn << "inserting " << armor64A (c) << "'s blocks\n";
      assert (_insert_this_commit == 0);
      _insert_this_commit = b;
      _insert_this_commit_hash = c;
      if (r->ndeltas () > 0) {
        _deltas.clear ();
        _delta_blocks.clear ();
        _delta_hashes.clear ();
        _ndeltas = 0;
        _deltas.setsize (r->ndeltas ());
        _delta_blocks.setsize (r->ndeltas ());
        _work++;
        get_deltas 
	  (r, wrap (this, &syncer::get_ind_cb, r),
	   wrap (this, &syncer::get_deltas_cb, r));
      }
    }

    if (_work == 0)
      check_commits ();
  }

  void
  check_commit (bool inlog, bigint c, ptr<ns_block> b)
  {
    if (b) {
      warn << armor64A (c) << ": in DHT\n";
      get_cached_commit (inlog, c, true, b);
    }
    else {
      warn << armor64A (c) << ": not in DHT\n";
      get_cached_commit (inlog, c, false, nsobj->retrieve_local (c));
    }
  }

  void
  check_commits_next ()
  {
    assert (_work == 0);
    assert (_err == 0);

    if (_inlogs.size () > 0) {
      bigint c = _inlogs.pop_front ();
      nsobj->retrieve_block
	(c, wrap (this, &syncer::check_commit, true, c), true);
    }
    else if (_depends.size () > 0) {
      bigint c = _depends.pop_front ();
      nsobj->retrieve_block
	(c, wrap (this, &syncer::check_commit, false, c), true);
    }
    else
      finalize ();
  }

  void
  insert_commit_done (ns_stat status, bigint n)
  {
    if (status) {
      warn << "cannot insert " << armor64A (n) << "\n";
      _err++;
      finalize ();
      return;
    }
    check_commits_next ();
  }

  void
  check_commits ()
  {
    assert (_work == 0);
    assert (_err == 0);

    if (_insert_this_commit) {
      ptr<ns_block> b = _insert_this_commit;
      _insert_this_commit = 0;
      bigint n = nsobj->insert_block
        (b->data, b->len, wrap (this, &syncer::insert_commit_done), true);
      assert (n == _insert_this_commit_hash);
    }
    else
      check_commits_next ();
  }

  void
  check_log (unsigned i, ptr<view_state> v, ptr<keyhash_payload> p)
  {
    _work--;
    
    lptr u = 0;
    if (!p || !(u = log_pointer::alloc (p->buf ()))) {
      warn << armor64A (v->commits.head (i))
	   << ": lptr not available on DHT\n";
      _newpks.push_back (v->commits.head (i));
      _stop [i] = 0;
    }
    else {
      if (p->version () < (long) v->pkv.seqn (i)) {
        warn << "stale pk " << armor64A (v->commits.head (i)) << "\n";
        _newpks.push_back (v->commits.head (i));
	_stop [i] = u->commit ();
      }
      else
	// log-head on DHT and replica the same, or replica has stale
	// version
	_stop [i] = v->commits.rptr (i);
    }

    if (_work == 0) {
      _commits = v->commits;
      for (unsigned i=0; i<v->commits.size (); i++) {
	if (v->commits.rptr(i) != 0 && 
	    (_ispush || v->commits.rptr (i) != _stop [i])) {
	  _inlogs.push_back (v->commits.rptr (i));
	  _done.insert (v->commits.rptr (i), true);
	}
      }
      check_commits ();
    }
  }

  void
  got_version (ptr<view_state> v, log_status status)
  { 
    if (status) {
      warn << "cannot load version\n";
      _err = 1;
      finalize ();
    }
    else {
      _stop.setsize (v->commits.size ());
      for (unsigned i=0; i<v->commits.size (); i++) {
        _work++;
        nsobj->retrieve_pk_from_net
	  (v->commits.head (i), wrap (this, &syncer::check_log, i, v));
      }
    }
  }

  void
  got_view (bigint vid, log_status status)
  {
    if (status) {
      warn << "cannot load project block\n";
      _err = 1;
      finalize ();
    }
    else {
      _vid = vid;
      // check if view is available on DHT
      nsobj->retrieve_pk_from_net (vid, wrap (this, &syncer::check_vid));
    }
  }

  void
  check_vid (ptr<keyhash_payload> p)
  {
    if (!p)
      _newpks.push_back (_vid);
    _view->fetch (wrap (this, &syncer::got_version));
  }

  void
  sync_connected (bigint c)
  {
    if (!nsobj->connected ()) {
      warn << "cannot connect to DHT, wait\n";
      sync_resched ();
      nsobj = 0;
      delete this;
      return;
    }
    _view = view::alloc (wrap (this, &syncer::got_view, c), c);
  }

public:
  syncer (bigint c, vec<str> n, bool push)
    : _ispush (push), _rlist (n)
  {
    _work = 0;
    _err = 0;
    _ndeltas = 0;
    _insert_this_commit = 0;

    nsobj = ns::alloc (c);
    nsobj->connect (wrap (this, &syncer::sync_connected, c));
  }
};

extern bool push_mode;

void
sync_repository (vec<str> n)
{
  nsobj = 0;

  if (n.size () == 0) {
    sync_resched ();
    return;
  }

  str p = n.pop_front ();
  bigint c;
  if (!dearmor64A_bigint (p, c)) {
    warn << "cannot parse " << p << "\n";
    sync_repository (n);
    return;
  }
  vNew syncer (c, n, push_mode);
}

